# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from .._entity._entity_ids import _EntityIds
from .._manager._manager import _Manager
from .._version._version_mixin import _VersionMixin
from ..common._utils import _Subscriber
from ..common.warn_if_inputs_not_ready import _warn_if_inputs_not_ready
from ..exceptions.exceptions import (
    InvalidSequence,
    InvalidSequenceId,
    ModelNotFound,
    NonExistingSequence,
    NonExistingTask,
    SequenceBelongsToNonExistingScenario,
)
from ..job._job_manager_factory import _JobManagerFactory
from ..job.job import Job
from ..notification import Event, EventEntityType, EventOperation, Notifier
from ..notification.event import _make_event
from ..reason import EntityDoesNotExist, EntityIsNotSubmittableEntity, ReasonCollection
from ..scenario._scenario_manager_factory import _ScenarioManagerFactory
from ..scenario.scenario import Scenario
from ..scenario.scenario_id import ScenarioId
from ..submission._submission_manager_factory import _SubmissionManagerFactory
from ..submission.submission import Submission
from ..task._task_manager_factory import _TaskManagerFactory
from ..task.task import Task, TaskId
from .sequence import Sequence
from .sequence_id import SequenceId


class _SequenceManager(_Manager[Sequence], _VersionMixin):
    _ENTITY_NAME = Sequence.__name__
    _EVENT_ENTITY_TYPE = EventEntityType.SEQUENCE
    _model_name = "sequences"

    @classmethod
    def _delete(cls, sequence_id: SequenceId) -> None:
        """
        Deletes a Sequence by id.
        """
        sequence_name, scenario_id = cls._breakdown_sequence_id(sequence_id)

        if scenario := _ScenarioManagerFactory._build_manager()._get(scenario_id):
            if sequence_name in scenario._sequences.keys():
                scenario.remove_sequences([sequence_name])
                if hasattr(cls, "_EVENT_ENTITY_TYPE"):
                    Notifier.publish(Event(cls._EVENT_ENTITY_TYPE, EventOperation.DELETION, entity_id=sequence_id))
                return
        raise ModelNotFound(cls._model_name, sequence_id)

    @classmethod
    def _delete_all(cls) -> None:
        """
        Deletes all Sequences.
        """
        scenarios = _ScenarioManagerFactory._build_manager()._get_all()
        for scenario in scenarios:
            scenario.sequences = {}
        if hasattr(cls, "_EVENT_ENTITY_TYPE"):
            Notifier.publish(Event(cls._EVENT_ENTITY_TYPE, EventOperation.DELETION, metadata={"delete_all": True}))

    @classmethod
    def _delete_many(cls, sequence_ids: Iterable[SequenceId]) -> None:
        """
        Deletes Sequence entities by a list of Sequence ids.
        """
        scenario_manager = _ScenarioManagerFactory._build_manager()

        scenario_ids_and_sequence_names_map: Dict[str, List[str]] = {}
        for sequence in sequence_ids:
            sequence_id = sequence.id if isinstance(sequence, Sequence) else sequence
            sequence_name, scenario_id = cls._breakdown_sequence_id(sequence_id)
            sequences_names = scenario_ids_and_sequence_names_map.get(scenario_id, [])
            sequences_names.append(sequence_name)
            scenario_ids_and_sequence_names_map[scenario_id] = sequences_names

        try:
            for scenario_id, sequence_names in scenario_ids_and_sequence_names_map.items():
                scenario = scenario_manager._get(scenario_id)
                for sequence_name in sequence_names:
                    del scenario._sequences[sequence_name]
                scenario_manager._set(scenario)

            if hasattr(cls, "_EVENT_ENTITY_TYPE"):
                for sequence_id in sequence_ids:
                    Notifier.publish(Event(cls._EVENT_ENTITY_TYPE, EventOperation.DELETION, entity_id=sequence_id))
        except (ModelNotFound, KeyError):
            cls.__log_error_entity_not_found(sequence_id)
            raise ModelNotFound(cls._model_name, sequence_id) from None

    @classmethod
    def _delete_by_version(cls, version_number: str) -> None:
        """
        Deletes Sequences by version number.
        """
        for scenario in _ScenarioManagerFactory()._build_manager()._repository._search("version", version_number):
            cls._delete_many(scenario.sequences.values())

    @classmethod
    def _hard_delete(cls, sequence_id: SequenceId) -> None:
        sequence = cls._get(sequence_id)
        entity_ids_to_delete = cls._get_children_entity_ids(sequence)
        entity_ids_to_delete.sequence_ids.add(sequence.id)
        cls._delete_entities_of_multiple_types(entity_ids_to_delete)

    @classmethod
    def _set(cls, sequence: Sequence) -> None:
        """
        Save or update a Sequence.
        """
        sequence_name, scenario_id = cls._breakdown_sequence_id(sequence.id)
        scenario_manager = _ScenarioManagerFactory._build_manager()

        if scenario := scenario_manager._get(scenario_id):
            sequence_data = {
                Scenario._SEQUENCE_TASKS_KEY: sequence._tasks,
                Scenario._SEQUENCE_SUBSCRIBERS_KEY: sequence._subscribers,
                Scenario._SEQUENCE_PROPERTIES_KEY: sequence._properties.data,
            }
            scenario._sequences[sequence_name] = sequence_data
            scenario_manager._set(scenario)
        else:
            cls._logger.error(f"Sequence {sequence.id} belongs to a non-existing Scenario {scenario_id}.")
            raise SequenceBelongsToNonExistingScenario(sequence.id, scenario_id)

    @staticmethod
    def __get_sequence_tasks(tasks: Union[List[Task], List[TaskId]]) -> List[Task]:
        task_manager = _TaskManagerFactory._build_manager()
        _tasks: List[Task] = []
        for task in tasks:
            if isinstance(task, Task):
                _tasks.append(task)
            elif _task := task_manager._get(task):
                _tasks.append(_task)
            else:
                raise NonExistingTask(task)
        return _tasks

    @classmethod
    def _build_sequence(
        cls,
        sequence_name: str,
        tasks: Union[List[Task], List[TaskId]],
        subscribers: Optional[List[_Subscriber]] = None,
        properties: Optional[Dict] = None,
        scenario_id: Optional[ScenarioId] = None,
        version: Optional[str] = None,
    ) -> Sequence:
        sequence_id = Sequence._new_id(sequence_name, scenario_id)
        _tasks = cls.__get_sequence_tasks(tasks)
        properties = properties if properties else {}
        properties["name"] = sequence_name
        version = version if version else cls._get_latest_version()
        return Sequence(
            properties=properties,
            tasks=_tasks,
            sequence_id=sequence_id,
            owner_id=scenario_id,
            parent_ids={scenario_id} if scenario_id else None,
            subscribers=subscribers,
            version=version,
        )

    @classmethod
    def _bulk_create_from_scenario(cls, scenario: Scenario) -> Dict[str, Sequence]:
        _sequences: Dict[str, Sequence] = {}

        for sequence_name, sequence_data in scenario._sequences.items():
            sequence = cls._create(
                sequence_name,
                sequence_data.get(scenario._SEQUENCE_TASKS_KEY, []),
                sequence_data.get(scenario._SEQUENCE_SUBSCRIBERS_KEY, []),
                sequence_data.get(scenario._SEQUENCE_PROPERTIES_KEY, {}),
                scenario.id,
                scenario.version,
            )
            if not isinstance(sequence, Sequence):
                raise NonExistingSequence(sequence_name, scenario.id)
            _sequences[sequence_name] = sequence

            Notifier.publish(_make_event(sequence, EventOperation.CREATION))

        return _sequences

    @classmethod
    def _create(
        cls,
        sequence_name: str,
        tasks: Union[List[Task], List[TaskId]],
        subscribers: Optional[List[_Subscriber]] = None,
        properties: Optional[Dict] = None,
        scenario_id: Optional[ScenarioId] = None,
        version: Optional[str] = None,
    ) -> Sequence:
        task_manager = _TaskManagerFactory._build_manager()
        _tasks = cls.__get_sequence_tasks(tasks)

        sequence = cls._build_sequence(sequence_name, _tasks, subscribers, properties, scenario_id, version)
        sequence_id = sequence.id

        for task in _tasks:
            if sequence_id not in task._parent_ids:
                task._parent_ids.update([sequence_id])
                task_manager._repository._save(task)

        if not sequence._is_consistent():
            raise InvalidSequence(sequence_id)

        return sequence

    @classmethod
    def _breakdown_sequence_id(cls, sequence_id: str) -> Tuple[str, str]:
        try:
            sequence_name, scenario_id = sequence_id.split(Scenario._ID_PREFIX)
            scenario_id = f"{Scenario._ID_PREFIX}{scenario_id}"
            sequence_name = sequence_name.split(Sequence._ID_PREFIX)[1].strip("_")
            sequence_name = sequence_name.replace("TPSPACE", " ")
            return sequence_name, scenario_id
        except (ValueError, IndexError):
            cls._logger.error(f"SequenceId {sequence_id} is invalid.")
            raise InvalidSequenceId(sequence_id) from None

    @classmethod
    def _get(cls, sequence: Union[str, Sequence], default=None) -> Sequence:
        """
        Returns a Sequence by id or reference.
        """
        try:
            sequence_id = sequence.id if isinstance(sequence, Sequence) else sequence
            sequence_name, scenario_id = cls._breakdown_sequence_id(sequence_id)

            scenario_manager = _ScenarioManagerFactory._build_manager()
            if scenario := scenario_manager._get(scenario_id):
                if sequence_entity := scenario.sequences.get(sequence_name, None):
                    return sequence_entity
            cls.__log_error_entity_not_found(sequence_id)
            return default
        except (ModelNotFound, InvalidSequenceId):
            cls.__log_error_entity_not_found(sequence_id)
            return default

    @classmethod
    def _get_all(cls, version_number: Optional[str] = None) -> List[Sequence]:
        """
        Returns all Sequence entities.
        """
        sequences = []
        scenarios = _ScenarioManagerFactory._build_manager()._get_all(version_number)
        for scenario in scenarios:
            sequences.extend(list(scenario.sequences.values()))
        return sequences

    @classmethod
    def _get_all_by(cls, filters: Optional[List[Dict]] = None) -> List[Sequence]:
        sequences = cls._get_all()

        if not filters:
            return sequences

        filtered_sequences = []
        for sequence in sequences:
            for filter in filters:
                if all(getattr(sequence, key) == item for key, item in filter.items()):
                    filtered_sequences.append(sequence)
        return filtered_sequences

    @classmethod
    def _get_children_entity_ids(cls, sequence: Sequence) -> _EntityIds:
        entity_ids = _EntityIds()
        for task in sequence.tasks.values():
            if not isinstance(task, Task):
                task = _TaskManagerFactory._build_manager()._get(task)
            if task.owner_id == sequence.id:
                entity_ids.task_ids.add(task.id)
            for data_node in task.data_nodes.values():
                if data_node.owner_id == sequence.id:
                    entity_ids.data_node_ids.add(data_node.id)

        jobs = _JobManagerFactory._build_manager()._get_all()
        for job in jobs:
            if job.task.id in entity_ids.task_ids:
                entity_ids.job_ids.add(job.id)

        submissions = _SubmissionManagerFactory._build_manager()._get_all()
        submitted_entity_ids = list(entity_ids.sequence_ids.union(entity_ids.task_ids))
        for submission in submissions:
            if submission.entity_id in submitted_entity_ids:
                entity_ids.submission_ids.add(submission.id)

        return entity_ids

    @classmethod
    def _subscribe(
        cls,
        callback: Callable[[Sequence, Job], None],
        params: Optional[List[Any]] = None,
        sequence: Optional[Sequence] = None,
    ) -> None:
        if sequence is None:
            sequences = cls._get_all()
            for pln in sequences:
                cls.__add_subscriber(callback, params, pln)
            return
        cls.__add_subscriber(callback, params, sequence)

    @classmethod
    def _unsubscribe(
        cls,
        callback: Callable[[Sequence, Job], None],
        params: Optional[List[Any]] = None,
        sequence: Optional[Sequence] = None,
    ) -> None:
        if sequence is None:
            sequences = cls._get_all()
            for pln in sequences:
                cls.__remove_subscriber(callback, params, pln)
            return
        cls.__remove_subscriber(callback, params, sequence)

    @classmethod
    def __add_subscriber(cls, callback, params, sequence):
        sequence._add_subscriber(callback, params)
        Notifier.publish(_make_event(sequence, EventOperation.UPDATE, attribute_name="subscribers"))

    @classmethod
    def __remove_subscriber(cls, callback, params, sequence):
        sequence._remove_subscriber(callback, params)
        Notifier.publish(_make_event(sequence, EventOperation.UPDATE, attribute_name="subscribers"))

    @classmethod
    def _is_submittable(cls, sequence: Union[Sequence, SequenceId]) -> ReasonCollection:
        reason_collector = ReasonCollection()

        if isinstance(sequence, str):
            sequence_id = sequence
            sequence = cls._get(sequence)
            if sequence is None:
                reason_collector._add_reason(sequence_id, EntityDoesNotExist(sequence_id))
                return reason_collector

        if not isinstance(sequence, Sequence):
            reason_collector._add_reason(str(sequence), EntityIsNotSubmittableEntity(str(sequence)))
        else:
            return sequence.is_ready_to_run()

        return reason_collector

    @classmethod
    def _submit(
        cls,
        sequence: Union[SequenceId, Sequence],
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        check_inputs_are_ready: bool = True,
        **properties,
    ) -> Submission:
        sequence_id = sequence.id if isinstance(sequence, Sequence) else sequence
        sequence = cls._get(sequence_id)
        if sequence is None:
            raise NonExistingSequence(sequence_id)
        callbacks = callbacks or []
        sequence_subscription_callback = cls.__get_status_notifier_callbacks(sequence) + callbacks
        if check_inputs_are_ready:
            _warn_if_inputs_not_ready(sequence.get_inputs())

        submission = (
            _TaskManagerFactory._build_manager()
            ._orchestrator()
            .submit(
                sequence,
                callbacks=sequence_subscription_callback,
                force=force,
                wait=wait,
                timeout=timeout,
                **properties,
            )
        )
        Notifier.publish(_make_event(sequence, EventOperation.SUBMISSION))
        return submission

    @classmethod
    def _exists(cls, entity_id: str) -> ReasonCollection:
        """
        Returns True if the entity id exists.
        """
        reason_collector = ReasonCollection()

        if cls._get(entity_id) is None:
            reason_collector._add_reason(entity_id, EntityDoesNotExist(entity_id))

        return reason_collector

    @classmethod
    def __log_error_entity_not_found(cls, sequence_id: Union[SequenceId, str]):
        cls._logger.error(f"{cls._ENTITY_NAME} not found: {str(sequence_id)}")

    @staticmethod
    def __get_status_notifier_callbacks(sequence: Sequence) -> List:
        return [partial(c.callback, *c.params, sequence) for c in sequence.subscribers]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import copy
import json
import pathlib
import shutil
from typing import Any, Dict, Iterable, Iterator, List, Optional, Type, Union

from taipy.common.config import Config

from ..common._utils import _retry_repository_operation
from ..common.typing import Converter, Entity, Json, ModelType
from ..exceptions import FileCannotBeRead, FileEmpty, ModelNotFound
from ._abstract_repository import _AbstractRepository
from ._decoder import _Decoder
from ._encoder import _Encoder


class _FileSystemRepository(_AbstractRepository[ModelType, Entity]):
    """
    Holds common methods to be used and extended when the need for saving
    dataclasses as JSON files in local storage emerges.

    Some lines have type: ignore because MyPy won't recognize some generic attributes. This
    should be revised in the future.

    Attributes:
        model_type (ModelType): Generic dataclass.
        converter: A class that handles conversion to and from a database backend
        dir_name (str): Folder that will hold the files for this dataclass model.
    """

    __EXCEPTIONS_TO_RETRY = (FileCannotBeRead, FileEmpty)

    def __init__(self, model_type: Type[ModelType], converter: Type[Converter], dir_name: str):
        self.model_type = model_type
        self.converter = converter
        self._dir_name = dir_name

    @property
    def dir_path(self):
        return self._storage_folder / self._dir_name

    @property
    def _storage_folder(self) -> pathlib.Path:
        return pathlib.Path(Config.core.taipy_storage_folder)

    ###############################
    # ##   Inherited methods   ## #
    ###############################

    def _save(self, entity: Entity):
        self.__create_directory_if_not_exists()
        model = self.converter._entity_to_model(entity)  # type: ignore
        self.__get_path(model.id).write_text(
            json.dumps(model.to_dict(), ensure_ascii=False, indent=0, cls=_Encoder, check_circular=False),
            encoding="UTF-8",
        )

    def _exists(self, entity_id: str) -> bool:
        return self.__get_path(entity_id).exists()

    def _load(self, entity_id: str) -> Entity:
        path = pathlib.Path(self.__get_path(entity_id))

        try:
            file_content = self.__read_file(path)
        except (FileNotFoundError, FileCannotBeRead, FileEmpty):
            raise ModelNotFound(str(self.dir_path), entity_id) from None

        return self.__file_content_to_entity(file_content)

    def _load_all(self, filters: Optional[List[Dict]] = None) -> List[Entity]:
        entities = []
        try:
            for f in self.dir_path.iterdir():
                if data := self.__filter_by(f, filters):
                    entities.append(self.__file_content_to_entity(data))
        except FileNotFoundError:
            pass
        return entities

    def _delete(self, entity_id: str):
        try:
            self.__get_path(entity_id).unlink()
        except FileNotFoundError:
            raise ModelNotFound(str(self.dir_path), entity_id) from None

    def _delete_all(self):
        shutil.rmtree(self.dir_path, ignore_errors=True)

    def _delete_many(self, ids: Iterable[str]):
        for model_id in ids:
            self._delete(model_id)

    def _delete_by(self, attribute: str, value: str):
        filters: List[Dict] = [{}]
        for fil in filters:
            fil.update({attribute: value})

        try:
            for f in self.dir_path.iterdir():
                if self.__filter_by(f, filters):
                    f.unlink()
        except FileNotFoundError:
            pass

    def _search(self, attribute: str, value: Any, filters: Optional[List[Dict]] = None) -> List[Entity]:
        return list(self.__search(attribute, value, filters))

    def _export(self, entity_id: str, folder_path: Union[str, pathlib.Path]) -> None:
        if isinstance(folder_path, str):
            folder: pathlib.Path = pathlib.Path(folder_path)
        else:
            folder = folder_path

        export_dir = folder / self._dir_name
        if not export_dir.exists():
            export_dir.mkdir(parents=True)

        export_path = export_dir / f"{entity_id}.json"

        shutil.copy2(self.__get_path(entity_id), export_path)

    ###########################################
    # ##   Specific or optimized methods   ## #
    ###########################################
    def _get_by_configs_and_owner_ids(self, configs_and_owner_ids, filters: Optional[List[Dict]] = None):
        # Design in order to optimize performance on Entity creation.
        # Maintainability and readability were impacted.
        if not filters:
            filters = [{}]
        res = {}
        configs_and_owner_ids = set(configs_and_owner_ids)

        try:
            for f in self.dir_path.iterdir():
                config_id, owner_id, entity = self.__match_file_and_get_entity(
                    f, configs_and_owner_ids, copy.deepcopy(filters)
                )

                if entity:
                    key = config_id, owner_id
                    res[key] = entity
                    configs_and_owner_ids.remove(key)

                    if len(configs_and_owner_ids) == 0:
                        return res
        except FileNotFoundError:
            # Folder with data was not created yet.
            return {}

        return res

    def _get_by_config_and_owner_id(
        self, config_id: str, owner_id: Optional[str], filters: Optional[List[Dict]] = None
    ) -> Optional[Entity]:
        filters = [{}] if not filters else copy.deepcopy(filters)

        if owner_id is not None:
            for fil in filters:
                fil.update({"owner_id": owner_id})
        return self.__filter_files_by_config_and_owner_id(config_id, owner_id, filters)

    #############################
    # ##   Private methods   ## #
    #############################

    def __filter_files_by_config_and_owner_id(
        self, config_id: str, owner_id: Optional[str], filters: Optional[List[Dict]] = None
    ):
        try:
            files = filter(lambda f: config_id in f.name, self.dir_path.iterdir())
            entities = (self.__file_content_to_entity(self.__filter_by(f, filters)) for f in files)
            corresponding_entities = filter(
                lambda e: e is not None and e.config_id == config_id and e.owner_id == owner_id,  # type: ignore
                entities,
            )
            return next(corresponding_entities, None)  # type: ignore
        except FileNotFoundError:
            pass
        return None

    def __match_file_and_get_entity(self, filepath, config_and_owner_ids, filters):
        if match := [(c, p) for c, p in config_and_owner_ids if c.id in filepath.name]:
            for config, owner_id in match:
                for fil in filters:
                    fil.update({"config_id": config.id, "owner_id": owner_id})

                if data := self.__filter_by(filepath, filters):
                    return config, owner_id, self.__file_content_to_entity(data)

        return None, None, None

    def __create_directory_if_not_exists(self):
        self.dir_path.mkdir(parents=True, exist_ok=True)

    def __search(self, attribute: str, value: str, filters: Optional[List[Dict]] = None) -> Iterator[Entity]:
        return filter(lambda e: getattr(e, attribute, None) == value, self._load_all(filters))

    def __get_path(self, model_id) -> pathlib.Path:
        return self.dir_path / f"{model_id}.json"

    def __file_content_to_entity(self, file_content):
        if not file_content:
            return None
        if isinstance(file_content, str):
            file_content = json.loads(file_content, cls=_Decoder)
        model = self.model_type.from_dict(file_content)
        return self.converter._model_to_entity(model)

    def __filter_by(self, filepath: pathlib.Path, filters: Optional[List[Dict]]) -> Optional[Json]:
        if not filters:
            filters = [{}]

        try:
            file_content = self.__read_file(filepath)
        except (FileNotFoundError, FileCannotBeRead, FileEmpty):
            return None

        for _filter in filters:
            conditions = [
                f'"{key}": "{value}"' if value is not None else f'"{key}": null' for key, value in _filter.items()
            ]
            if all(condition in file_content for condition in conditions):
                return json.loads(file_content, cls=_Decoder)
        return None

    @_retry_repository_operation(__EXCEPTIONS_TO_RETRY)
    def __read_file(self, filepath: pathlib.Path) -> str:
        if not filepath.is_file():
            raise FileNotFoundError

        try:
            with filepath.open("r", encoding="UTF-8") as f:
                file_content = f.read()
            if not file_content:
                raise FileEmpty(str(filepath))
            return file_content
        except Exception:
            raise FileCannotBeRead(str(filepath)) from None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import functools
import threading

from .._manager._manager import _Manager
from ..notification import EventOperation, Notifier, _make_event


class _Reloader:
    """The _Reloader singleton class"""

    _instance = None
    _lock = threading.RLock()

    def __new__(cls):
        with cls._lock:
            if not isinstance(cls._instance, cls):
                cls._instance = super().__new__(cls)
                cls._instance._no_reload_context = False  # Initialize once
                cls._instance._context_depth = 0  # Track nested `with` usage
        return cls._instance

    def _reload(self, manager: str, obj):
        if self._no_reload_context:
            return obj

        entity = _get_manager(manager)._get(obj, obj)
        if obj._is_in_context and hasattr(entity, "_properties"):
            if obj._properties._pending_changes:
                entity._properties._pending_changes = obj._properties._pending_changes
            if obj._properties._pending_deletions:
                entity._properties._pending_deletions = obj._properties._pending_deletions
            entity._properties._entity_owner = obj
        return entity

    def __enter__(self):
        with self._lock:
            self._context_depth += 1
            self._no_reload_context = True
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        with self._lock:
            self._context_depth -= 1
            if self._context_depth == 0:
                self._no_reload_context = False


def _self_reload(manager: str):
    def __reload(fct):
        @functools.wraps(fct)
        def _do_reload(self, *args, **kwargs):
            self = _Reloader()._reload(manager, self)
            return fct(self, *args, **kwargs)

        return _do_reload

    return __reload


def _self_setter(manager):
    def __set_entity(fct):
        @functools.wraps(fct)
        def _do_set_entity(self, *args, **kwargs):
            fct(self, *args, **kwargs)
            entity_manager = _get_manager(manager)
            value = args[0] if len(args) == 1 else args
            event = _make_event(
                self,
                EventOperation.UPDATE,
                attribute_name=fct.__name__,
                attribute_value=value,
            )
            if not self._is_in_context:
                entity = _Reloader()._reload(manager, self)
                fct(entity, *args, **kwargs)
                entity_manager._set(entity)
                Notifier.publish(event)
            else:
                self._in_context_attributes_changed_collector.append(event)

        return _do_set_entity

    return __set_entity


@functools.lru_cache
def _get_manager(manager: str) -> _Manager:
    from ..cycle._cycle_manager_factory import _CycleManagerFactory
    from ..data._data_manager_factory import _DataManagerFactory
    from ..job._job_manager_factory import _JobManagerFactory
    from ..scenario._scenario_manager_factory import _ScenarioManagerFactory
    from ..sequence._sequence_manager_factory import _SequenceManagerFactory
    from ..submission._submission_manager_factory import _SubmissionManagerFactory
    from ..task._task_manager_factory import _TaskManagerFactory

    return {
        "scenario": _ScenarioManagerFactory._build_manager(),
        "sequence": _SequenceManagerFactory._build_manager(),
        "data": _DataManagerFactory._build_manager(),
        "cycle": _CycleManagerFactory._build_manager(),
        "job": _JobManagerFactory._build_manager(),
        "task": _TaskManagerFactory._build_manager(),
        "submission": _SubmissionManagerFactory._build_manager(),
    }[manager]  # type: ignore



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Union

from .._entity._entity import _Entity
from ..cycle.cycle import Cycle
from ..data.data_node import DataNode
from ..job.job import Job
from ..scenario.scenario import Scenario
from ..sequence.sequence import Sequence
from ..submission.submission import Submission
from ..task.task import Task


def _is_cycle(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Cycle) or (isinstance(entity, str) and entity.startswith(Cycle._ID_PREFIX))


def _is_scenario(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Scenario) or (isinstance(entity, str) and entity.startswith(Scenario._ID_PREFIX))


def _is_sequence(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Sequence) or (isinstance(entity, str) and entity.startswith(Sequence._ID_PREFIX))


def _is_task(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Task) or (isinstance(entity, str) and entity.startswith(Task._ID_PREFIX))


def _is_job(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Job) or (isinstance(entity, str) and entity.startswith(Job._ID_PREFIX))


def _is_data_node(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, DataNode) or (isinstance(entity, str) and entity.startswith(DataNode._ID_PREFIX))


def _is_submission(entity: Union[_Entity, str]) -> bool:
    return isinstance(entity, Submission) or (isinstance(entity, str) and entity.startswith(Submission._ID_PREFIX))



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from .._repository._base_taipy_model import _BaseModel
from ..job.job_id import JobId
from .submission_status import SubmissionStatus


@dataclass
class _SubmissionModel(_BaseModel):
    id: str
    entity_id: str
    entity_type: str
    entity_config_id: Optional[str]
    job_ids: Union[List[JobId], List]
    properties: Dict[str, Any]
    creation_date: str
    submission_status: SubmissionStatus
    version: str
    is_completed: bool
    is_abandoned: bool
    is_canceled: bool
    running_jobs: List[str]
    blocked_jobs: List[str]
    pending_jobs: List[str]

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _SubmissionModel(
            id=data["id"],
            entity_id=data["entity_id"],
            entity_type=data["entity_type"],
            entity_config_id=data.get("entity_config_id"),
            job_ids=_BaseModel._deserialize_attribute(data["job_ids"]),
            properties=_BaseModel._deserialize_attribute(data["properties"]),
            creation_date=data["creation_date"],
            submission_status=SubmissionStatus._from_repr(data["submission_status"]),
            version=data["version"],
            is_completed=data["is_completed"],
            is_abandoned=data["is_abandoned"],
            is_canceled=data["is_canceled"],
            running_jobs=_BaseModel._deserialize_attribute(data["running_jobs"]),
            blocked_jobs=_BaseModel._deserialize_attribute(data["blocked_jobs"]),
            pending_jobs=_BaseModel._deserialize_attribute(data["pending_jobs"]),
        )

    def to_list(self):
        return [
            self.id,
            self.entity_id,
            self.entity_type,
            self.entity_config_id,
            _BaseModel._serialize_attribute(self.job_ids),
            _BaseModel._serialize_attribute(self.properties),
            self.creation_date,
            repr(self.submission_status),
            self.version,
            self.is_completed,
            self.is_abandoned,
            self.is_canceled,
            _BaseModel._serialize_attribute(self.running_jobs),
            _BaseModel._serialize_attribute(self.blocked_jobs),
            _BaseModel._serialize_attribute(self.pending_jobs),
        ]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any, Callable, Dict, List, Union

import numpy as np
import pandas as pd

from ..exceptions.exceptions import InvalidExposedType


class _TabularDataNodeMixin(object):
    """Mixin class designed to handle tabular representable data nodes."""

    _HAS_HEADER_PROPERTY = "has_header"
    _EXPOSED_TYPE_PROPERTY = "exposed_type"
    _EXPOSED_TYPE_NUMPY = "numpy"
    _EXPOSED_TYPE_PANDAS = "pandas"
    _EXPOSED_TYPE_MODIN = "modin"  # Deprecated in favor of pandas since 3.1.0
    _VALID_STRING_EXPOSED_TYPES = [_EXPOSED_TYPE_PANDAS, _EXPOSED_TYPE_NUMPY]

    def __init__(self, **kwargs) -> None:
        self._decoder: Union[Callable, Any]
        self.custom_document = kwargs.get(self._EXPOSED_TYPE_PROPERTY)

        custom_decoder = getattr(self.custom_document, "decode", None)
        if callable(custom_decoder):
            self._decoder = custom_decoder
        elif kwargs.get(self._HAS_HEADER_PROPERTY, True):
            self._decoder = self._default_decoder_with_header
        else:
            self._decoder = self._default_decoder_without_header

        self._encoder = self._default_encoder
        custom_encoder = getattr(self.custom_document, "encode", None)
        if callable(custom_encoder):
            self._encoder = custom_encoder


    def _convert_data_to_dataframe(self, exposed_type: Any, data: Any) -> Union[pd.DataFrame, pd.Series]:
        if exposed_type == self._EXPOSED_TYPE_PANDAS and isinstance(data, (pd.DataFrame, pd.Series)):
            return data
        elif exposed_type == self._EXPOSED_TYPE_NUMPY and isinstance(data, np.ndarray):
            return pd.DataFrame(data)
        elif isinstance(data, list) and not isinstance(exposed_type, str):
            return pd.DataFrame.from_records([self._encoder(row) for row in data])
        return pd.DataFrame(data)

    @classmethod
    def _get_valid_exposed_type(cls, properties: Dict):
        if (
            cls._EXPOSED_TYPE_PROPERTY not in properties.keys()
            or properties[cls._EXPOSED_TYPE_PROPERTY] == cls._EXPOSED_TYPE_MODIN
        ):
            # Default exposed type is pandas
            # Deprecated modin exposed type in favor of pandas since 3.1.0
            return cls._EXPOSED_TYPE_PANDAS
        return properties[cls._EXPOSED_TYPE_PROPERTY]

    @classmethod
    def _check_exposed_type(cls, exposed_type):
        valid_string_exposed_types = cls._VALID_STRING_EXPOSED_TYPES
        if isinstance(exposed_type, str) and exposed_type not in valid_string_exposed_types:
            raise InvalidExposedType(
                f"Invalid string exposed type {exposed_type}. Supported values are "
                f"{', '.join(valid_string_exposed_types)}"
            )

    def _default_decoder_with_header(self, document: Dict) -> Any:
        if self.custom_document:
            return self.custom_document(**document)

    def _default_decoder_without_header(self, document: List) -> Any:
        if self.custom_document:
            return self.custom_document(*document)

    def _default_encoder(self, document_object: Any) -> Dict:
        return document_object.__dict__



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import pathlib
import shutil
from datetime import datetime
from os.path import isfile
from typing import Any, Callable, Dict, Optional, cast

from taipy.common.config import Config
from taipy.common.logger._taipy_logger import _TaipyLogger

from .._entity._reload import _self_reload
from ..common._utils import _normalize_path
from ..reason import InvalidUploadFile, NoFileToDownload, NotAFile, ReasonCollection, UploadFileCanNotBeRead
from .data_node import DataNode
from .data_node_id import Edit


class _FileDataNodeMixin(object):
    """Mixin class designed to handle file-based data nodes."""

    __EXTENSION_MAP = {"csv": "csv", "excel": "xlsx", "parquet": "parquet", "pickle": "p", "json": "json"}

    _DEFAULT_DATA_KEY = "default_data"
    _PATH_KEY = "path"
    _DEFAULT_PATH_KEY = "default_path"
    _IS_GENERATED_KEY = "is_generated"

    __logger = _TaipyLogger._get_logger()

    def __init__(self, properties: Dict) -> None:
        self._path: str = cast(str, properties.get(self._PATH_KEY, properties.get(self._DEFAULT_PATH_KEY)))
        self._is_generated: bool = properties.get(self._IS_GENERATED_KEY, self._path is None)
        self._last_edit_date: Optional[datetime] = None

        if self._path and ".data" in self._path:
            self._path = self._migrate_path(self.storage_type(), self._path)  # type: ignore[attr-defined]
        if not self._path:
            self._path = self._build_path(self.storage_type())  # type: ignore[attr-defined]

        properties[self._IS_GENERATED_KEY] = self._is_generated
        properties[self._PATH_KEY] = self._path

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def is_generated(self) -> bool:
        """Indicates if the file is generated."""
        return self._is_generated

    @property  # type: ignore
    @_self_reload(DataNode._MANAGER_NAME)
    def path(self) -> str:
        """The path to the file data of the data node."""
        return _normalize_path(self._path)

    @path.setter
    def path(self, value) -> None:
        _path = _normalize_path(value)
        self._path = _path
        self.properties[self._PATH_KEY] = _path  # type: ignore[attr-defined]
        self.properties[self._IS_GENERATED_KEY] = False  # type: ignore[attr-defined]

    def is_downloadable(self) -> ReasonCollection:
        """Check if the data node is downloadable.

        Returns:
            A `ReasonCollection^` object containing the reasons why the data node is not downloadable.
        """
        collection = ReasonCollection()
        if not os.path.exists(self.path):
            collection._add_reason(self.id, NoFileToDownload(self.path, self.id))  # type: ignore[attr-defined]
        elif not isfile(self.path):
            collection._add_reason(self.id, NotAFile(self.path, self.id))  # type: ignore[attr-defined]
        return collection

    def is_uploadable(self) -> ReasonCollection:
        """Check if the data node is upload-able.

        Returns:
            A `ReasonCollection^` object containing the reasons why the data node is not upload-able.
        """
        return ReasonCollection()

    def _get_downloadable_path(self) -> str:
        """Get the downloadable path of the file data of the data node.

        Returns:
            The downloadable path of the file data of the data node if it exists, otherwise an empty string.
        """
        if os.path.exists(self.path) and isfile(self._path):
            return self.path

        return ""

    def _upload(self, path: str, upload_checker: Optional[Callable[[str, Any], bool]] = None) -> ReasonCollection:
        """Upload a file data to the data node.

        Parameters:
            path (str): The path of the file to upload to the data node.
            upload_checker (Optional[Callable[[str, Any], bool]]): A function to check if the upload is allowed.
                The function takes the title of the upload data and the data itself as arguments and returns
                True if the upload is allowed, otherwise False.

        Returns:
            True if the upload was successful, otherwise False.
        """
        from ._data_manager_factory import _DataManagerFactory

        reason_collection = ReasonCollection()

        upload_path = pathlib.Path(path)

        try:
            upload_data = self._read_from_path(str(upload_path))
        except Exception as err:
            self.__logger.error(f"Error while uploading {upload_path.name} to data node {self.id}:")  # type: ignore[attr-defined]
            self.__logger.error(f"Error: {err}")
            reason_collection._add_reason(self.id, UploadFileCanNotBeRead(upload_path.name, self.id))  # type: ignore[attr-defined]
            return reason_collection

        if upload_checker is not None:
            try:
                can_upload = upload_checker(upload_path.name, upload_data)
            except Exception as err:
                self.__logger.error(
                    f"Error while checking if {upload_path.name} can be uploaded to data node {self.id}"  # type: ignore[attr-defined]
                    f" using the upload checker {upload_checker.__name__}: {err}"
                )
                can_upload = False

            if not can_upload:
                reason_collection._add_reason(self.id, InvalidUploadFile(upload_path.name, self.id))  # type: ignore[attr-defined]
                return reason_collection

        shutil.copy(upload_path, self.path)

        self.track_edit(timestamp=datetime.now())  # type: ignore[attr-defined]
        self.unlock_edit()  # type: ignore[attr-defined]
        _DataManagerFactory._build_manager()._set(self)  # type: ignore[arg-type]

        return reason_collection

    def _read_from_path(self, path: Optional[str] = None, **read_kwargs) -> Any:
        raise NotImplementedError

    def _write_default_data(self, default_value: Any):
        if default_value is not None and not os.path.exists(self._path):
            self._write(default_value)  # type: ignore[attr-defined]
            self._last_edit_date = DataNode._get_last_modified_datetime(self._path) or datetime.now()
            self._edits.append(  # type: ignore[attr-defined]
                Edit(
                    {
                        "timestamp": self._last_edit_date,
                        "editor": "TAIPY",
                        "comment": "Default data written.",
                    }
                )
            )

        if not self._last_edit_date and isfile(self._path):
            self._last_edit_date = datetime.now()

    def _build_path(self, storage_type) -> str:
        folder = f"{storage_type}s"
        dir_path = pathlib.Path(Config.core.storage_folder) / folder
        if not dir_path.exists():
            dir_path.mkdir(parents=True, exist_ok=True)
        return str(dir_path / f"{self.id}.{self.__EXTENSION_MAP.get(storage_type)}")  # type: ignore[attr-defined]

    def _migrate_path(self, storage_type, old_path) -> str:
        new_path = self._build_path(storage_type)
        if os.path.exists(old_path):
            shutil.move(old_path, new_path)
        return new_path



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import uuid
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Union

from taipy.common.config.common._validate_id import _validate_id
from taipy.common.config.common.scope import Scope

from .._entity._entity import _Entity
from .._entity._labeled import _Labeled
from .._entity._properties import _Properties
from .._entity._reload import _Reloader, _self_reload, _self_setter
from .._version._version_manager_factory import _VersionManagerFactory
from ..data.data_node import DataNode
from ..exceptions import AttributeKeyAlreadyExisted
from ..notification.event import Event, EventEntityType, EventOperation, _make_event
from ..submission.submission import Submission
from .task_id import TaskId


class Task(_Entity, _Labeled):
    """Hold a user function that will be executed, its parameters and the results.

    A `Task` brings together the user code as function, the inputs and the outputs
    as data nodes (instances of the `DataNode^` class).

    !!! note
        It is not recommended to instantiate a `Task` directly. Instead, it should be
        created with the `create_scenario()^` function. When creating a `Scenario^`,
        the related data nodes and tasks are created automatically. Please refer to
        the `Scenario^` class for more information.

    A task's attributes (the input data nodes, the output data nodes, the Python
    function) are populated based on its task configuration `TaskConfig^`.

    ??? Example

        ```python
        import taipy as tp
        from taipy import Config

        def by_two(x: int):
            return x * 2

        if __name__ == "__main__":
            # Configure data nodes, tasks and scenarios
            input_cfg = Config.configure_data_node("my_input", default_data=2)
            result_cfg = Config.configure_data_node("my_result")
            task_cfg = Config.configure_task("my_double", function=by_two, input=input_cfg, output=result_cfg)
            scenario_cfg = Config.configure_scenario("my_scenario", task_configs=[task_cfg])

            # Instantiate a task along with a scenario
            sc = tp.create_scenario(scenario_cfg)

            # Retrieve task and data nodes from scenario
            task_input = sc.my_input
            double_task = sc.my_double
            task_result = sc.my_result

            # Write the input data and submit the task
            task_input.write(3)
            double_task.submit()

            # Read the result
            print(task_result.read())  # Output: 6

            # Retrieve the list of all tasks
            all_tasks = tp.get_tasks()
        ```

    Attributes:
        config_id (str): The identifier of the `TaskConfig^`.
        properties (dict[str, Any]): A dictionary of additional properties.
        function (callable): The python function to execute. The _function_ must take as parameter the
            data referenced by inputs data nodes, and must return the data referenced by outputs data nodes.
        input (Union[DataNode^, List[DataNode^]]): The list of inputs.
        output (Union[DataNode^, List[DataNode^]]): The list of outputs.
        id (str): The unique identifier of the task.
        owner_id (str):  The identifier of the owner (sequence_id, scenario_id, cycle_id) or None.
        parent_ids (Optional[Set[str]]): The set of identifiers of the parent sequences.
        version (str): The string indicates the application version of the task to instantiate. If not provided, the
            latest version is used.
        skippable (bool): If True, indicates that the task can be skipped if no change has been made on inputs. The
            default value is _False_.

    """

    _ID_PREFIX = "TASK"
    __ID_SEPARATOR = "_"
    _MANAGER_NAME = "task"
    __CHECK_INIT_DONE_ATTR_NAME = "_init_done"

    id: TaskId
    """The unique identifier of the task."""

    def __init__(
        self,
        config_id: str,
        properties: Dict[str, Any],
        function: Callable,
        input: Optional[Iterable[DataNode]] = None,
        output: Optional[Iterable[DataNode]] = None,
        id: Optional[TaskId] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        version: Optional[str] = None,
        skippable: bool = False,
    ) -> None:
        self._config_id = _validate_id(config_id)
        self.id = id or TaskId(self.__ID_SEPARATOR.join([self._ID_PREFIX, self.config_id, str(uuid.uuid4())]))
        self._owner_id = owner_id
        self._parent_ids = parent_ids or set()
        self._input = {dn.config_id: dn for dn in input or []}
        self._output = {dn.config_id: dn for dn in output or []}
        self._function = function
        self._version = version or _VersionManagerFactory._build_manager()._get_latest_version()
        self._skippable = skippable
        self._properties = _Properties(self, **properties)
        self._init_done = True

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other) -> bool:
        """Check if a task is equal to another task."""
        return isinstance(other, Task) and self.id == other.id

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, state):
        vars(self).update(state)

    def __setattr__(self, name: str, value: Any) -> None:
        if self.__CHECK_INIT_DONE_ATTR_NAME not in dir(self) or name in dir(self):
            return super().__setattr__(name, value)
        else:
            try:
                self.__getattr__(name)
                raise AttributeKeyAlreadyExisted(name)
            except AttributeError:
                return super().__setattr__(name, value)

    def __getattr__(self, attribute_name) -> Any:
        protected_attribute_name = _validate_id(attribute_name)
        if protected_attribute_name in self.input:
            return self.input[protected_attribute_name]
        if protected_attribute_name in self.output:
            return self.output[protected_attribute_name]
        raise AttributeError(f"{attribute_name} is not an attribute of task {self.id}")

    @property
    def properties(self) -> _Properties:
        """Dictionary of additional properties."""
        self._properties = _Reloader()._reload(self._MANAGER_NAME, self)._properties
        return self._properties

    @property
    def config_id(self) -> str:
        """The identifier of the `TaskConfig^`."""
        return self._config_id

    @property
    def owner_id(self) -> Optional[str]:
        """The identifier of the owner (scenario_id or cycle_id) or None."""
        return self._owner_id

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def parent_ids(self) -> Set[str]:
        """The set of identifiers of the parent scenarios."""
        return self._parent_ids

    @property
    def input(self) -> Dict[str, DataNode]:
        """The dictionary of input data nodes."""
        return self._input

    @property
    def output(self) -> Dict[str, DataNode]:
        """The dictionary of output data nodes."""
        return self._output

    @property
    def data_nodes(self) -> Dict[str, DataNode]:
        """The dictionary of input and output data nodes."""
        return {**self.input, **self.output}

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def function(self) -> Callable:
        """The python function to execute."""
        return self._function

    @function.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def function(self, val) -> None:
        self._function = val

    @property  # type: ignore
    @_self_reload(_MANAGER_NAME)
    def skippable(self) -> bool:
        """True if the task can be skipped if no change has been made on inputs. False otherwise"""
        return self._skippable

    @skippable.setter  # type: ignore
    @_self_setter(_MANAGER_NAME)
    def skippable(self, val) -> None:
        self._skippable = val

    @property
    def scope(self) -> Scope:
        """The lowest scope of the task's data nodes.

        The lowest scope present in input and output data nodes or GLOBAL if there are
        either no input or no output.
        """
        data_nodes = list(self._input.values()) + list(self._output.values())
        return Scope(min(dn.scope for dn in data_nodes)) if len(data_nodes) != 0 else Scope.GLOBAL

    @property
    def version(self) -> str:
        """The application version of the task.

        The string indicates the application version of the task. If not
        provided, the latest version is used."""
        return self._version

    def submit(
        self,
        callbacks: Optional[List[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        **properties,
    ) -> Submission:
        """Submit the task for execution.

        Parameters:
            callbacks (List[Callable]): The list of callable functions to be called on status
                change.
            force (bool): Force execution even if the data nodes are in cache.
            wait (bool): Wait for the orchestrated job created from the task submission to be finished in asynchronous
                mode.
            timeout (Union[float, int]): The maximum number of seconds to wait for the job to be finished before
                returning.<br/>
                If not provided and *wait* is True, the function waits indefinitely.
            **properties (dict[str, any]): A keyworded variable length list of additional arguments.

        Returns:
            A `Submission^` containing the information of the submission.
        """
        from ._task_manager_factory import _TaskManagerFactory

        return _TaskManagerFactory._build_manager()._submit(self, callbacks, force, wait, timeout, **properties)

    def get_parents(self) -> Dict[str, Set[_Entity]]:
        """Get the parent scenarios of the task.

        Returns:
            The dictionary of all parent entities.
                They are grouped by their type (Scenario^, Sequences^, or tasks^) so each key corresponds
                to a level of the parents and the value is a set of the parent entities.
                An empty dictionary is returned if the entity does not have parents.
        """
        from ... import core as tp

        return tp.get_parents(self)

    def get_label(self) -> str:
        """Returns the task simple label prefixed by its owner label.

        Returns:
            The label of the task as a string.
        """
        return self._get_label()

    def get_simple_label(self) -> str:
        """Returns the task simple label.

        Returns:
            The simple label of the task as a string.
        """
        return self._get_simple_label()


@_make_event.register(Task)
def _make_event_for_task(
    task: Task,
    operation: EventOperation,
    /,
    attribute_name: Optional[str] = None,
    attribute_value: Optional[Any] = None,
    **kwargs,
) -> Event:
    metadata = {"version": task.version, "config_id": task.config_id, **kwargs}
    return Event(
        entity_type=EventEntityType.TASK,
        entity_id=task.id,
        operation=operation,
        attribute_name=attribute_name,
        attribute_value=attribute_value,
        metadata=metadata,
    )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import sys
from typing import List

from taipy.common._cli._base_cli._abstract_cli import _AbstractCLI
from taipy.common._cli._base_cli._taipy_parser import _TaipyParser
from taipy.common.config import Config

from ._migrate import (
    _migrate_fs_entities,
    _migrate_mongo_entities,
    _remove_backup_file_entities,
    _remove_backup_mongo_entities,
    _restore_migrate_file_entities,
    _restore_migrate_mongo_entities,
)


class _MigrateCLI(_AbstractCLI):
    _COMMAND_NAME = "migrate"
    _ARGUMENTS = ["--repository-type", "--skip-backup", "--restore", "--remove-backup"]

    @classmethod
    def create_parser(cls):
        migrate_parser = _TaipyParser._add_subparser(
            cls._COMMAND_NAME,
            help="Migrate entities created from old taipy versions to be compatible with the current taipy version. "
            " The entity migration should be performed only after updating taipy code to the current version.",
        )
        migrate_parser.add_argument(
            "--repository-type",
            nargs="+",
            help="The type of repository to migrate. If filesystem, a path to the database folder should be informed. "
            "In case of mongo host, port, user and password must be informed, if left empty it "
            "is assumed default values",
        )
        migrate_parser.add_argument(
            "--skip-backup",
            action="store_true",
            help="Skip the backup of entities before migration.",
        )
        migrate_parser.add_argument(
            "--restore",
            action="store_true",
            help="Restore the migration of entities from backup folder.",
        )
        migrate_parser.add_argument(
            "--remove-backup",
            action="store_true",
            help="Remove the backup of entities. Only use this option if the migration was successful.",
        )

    @classmethod
    def handle_command(cls):
        args = cls._parse_arguments()
        if not args:
            return

        if not args.repository_type:
            _TaipyParser._sub_taipyparsers.get(cls._COMMAND_NAME).print_help()
            cls._logger.error("The following arguments are required: --repository-type")
            sys.exit(1)

        repository_type = args.repository_type[0]
        repository_args = args.repository_type[1:] if len(args.repository_type) > 1 else [None]

        if args.restore:
            cls.__handle_restore_backup(repository_type, repository_args)
        if args.remove_backup:
            cls.__handle_remove_backup(repository_type, repository_args)

        do_backup = not args.skip_backup
        cls.__migrate_entities(repository_type, repository_args, do_backup)
        sys.exit(0)

    @classmethod
    def __handle_remove_backup(cls, repository_type: str, repository_args: List):
        if repository_type == "filesystem":
            path = repository_args[0] or Config.core.taipy_storage_folder
            if not _remove_backup_file_entities(path):
                sys.exit(1)
        elif repository_type == "mongo":
            if not _remove_backup_mongo_entities():
                sys.exit(1)
        else:
            cls._logger.error(f"Unknown repository type {repository_type}")
            sys.exit(1)

        sys.exit(0)

    @classmethod
    def __handle_restore_backup(cls, repository_type: str, repository_args: List):
        if repository_type == "filesystem":
            path = repository_args[0] or Config.core.taipy_storage_folder
            if not _restore_migrate_file_entities(path):
                sys.exit(1)
        elif repository_type == "mongo":
            mongo_args = repository_args[1:5] if repository_args[0] else []
            if not _restore_migrate_mongo_entities(*mongo_args):
                sys.exit(1)
        else:
            cls._logger.error(f"Unknown repository type {repository_type}")
            sys.exit(1)
        sys.exit(0)

    @classmethod
    def __migrate_entities(cls, repository_type: str, repository_args: List, do_backup: bool):
        if repository_type == "filesystem":
            path = repository_args[0] or Config.core.taipy_storage_folder
            if not _migrate_fs_entities(path, do_backup):
                sys.exit(1)

        elif repository_type == "mongo":
            mongo_args = repository_args[1:5] if repository_args[0] else []
            _migrate_mongo_entities(*mongo_args, backup=do_backup)  # type: ignore

        else:
            cls._logger.error(f"Unknown repository type {repository_type}")
            sys.exit(1)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set, Union

import pandas as pd
from sqlalchemy import MetaData, Table

from taipy.common.config.common.scope import Scope

from .._version._version_manager_factory import _VersionManagerFactory
from ..exceptions.exceptions import MissingRequiredProperty
from ._abstract_sql import _AbstractSQLDataNode
from .data_node_id import DataNodeId, Edit


class SQLTableDataNode(_AbstractSQLDataNode):
    """Data Node stored in a SQL table.
    The *properties* attribute must contain the following mandatory entries:

    - *has_header* (`bool`): If True, indicates that the SQL query has a header.
    - *exposed_type* (`str`): The exposed type of the data read from SQL query. The default value is `pandas`.
    - *db_name* (`str`): The database name, or the name of the SQLite database file.
    - *db_engine* (`str`): The database engine. Possible values are *sqlite*, *mssql*,
        *mysql*, or *postgresql*.
    - *table_name* (`str`): The name of the SQL table.
    - *db_username* (`str`): The database username.
    - *db_password* (`str`): The database password.
    - *db_host* (`str`): The database host. The default value is *localhost*.
    - *db_port* (`int`): The database port. The default value is 1433.
    - *db_driver* (`str`): The database driver.

    The *properties* attribute can also contain the following optional entries:

    - *sqlite_folder_path* (str): The path to the folder that contains SQLite file. The default value
        is the current working folder.
    - *sqlite_file_extension* (str): The filename extension of the SQLite file. The default value is ".db".
    - *db_extra_args* (`Dict[str, Any]`): A dictionary of additional arguments to be passed into database
        connection string.
    """

    __STORAGE_TYPE = "sql_table"
    __TABLE_KEY = "table_name"

    def __init__(
        self,
        config_id: str,
        scope: Scope,
        id: Optional[DataNodeId] = None,
        owner_id: Optional[str] = None,
        parent_ids: Optional[Set[str]] = None,
        last_edit_date: Optional[datetime] = None,
        edits: Optional[List[Edit]] = None,
        version: Optional[str] = None,
        validity_period: Optional[timedelta] = None,
        edit_in_progress: bool = False,
        editor_id: Optional[str] = None,
        editor_expiration_date: Optional[datetime] = None,
        properties: Optional[Dict] = None,
    ) -> None:
        if properties is None:
            properties = {}
        if properties.get(self.__TABLE_KEY) is None:
            raise MissingRequiredProperty(f"Property {self.__TABLE_KEY} is not informed and is required.")
        super().__init__(
            config_id,
            scope,
            id=id,
            owner_id=owner_id,
            parent_ids=parent_ids,
            last_edit_date=last_edit_date,
            edits=edits,
            version=version or _VersionManagerFactory._build_manager()._get_latest_version(),
            validity_period=validity_period,
            edit_in_progress=edit_in_progress,
            editor_id=editor_id,
            editor_expiration_date=editor_expiration_date,
            properties=properties,
        )
        self._TAIPY_PROPERTIES.update({self.__TABLE_KEY})

    @classmethod
    def storage_type(cls) -> str:
        """Return the storage type of the data node: `sql_table`."""
        return cls.__STORAGE_TYPE

    def _get_base_read_query(self) -> str:
        return f"SELECT * FROM {self.properties[self.__TABLE_KEY]}"

    def _do_append(self, data, engine, connection) -> None:
        self.__insert_data(data, engine, connection)

    def _do_write(self, data, engine, connection) -> None:
        self.__insert_data(data, engine, connection, delete_table=True)

    def __insert_data(self, data, engine, connection, delete_table: bool = False) -> None:
        table = self._create_table(engine)
        self._insert_dataframe(
            self._convert_data_to_dataframe(self.properties[self._EXPOSED_TYPE_PROPERTY], data),
            table,
            connection,
            delete_table,
        )

    def _create_table(self, engine) -> Table:
        return Table(
            self.properties[self.__TABLE_KEY],
            MetaData(),
            autoload_with=engine,
        )

    @classmethod
    def _insert_dicts(cls, data: List[Dict], table: Any, connection: Any, delete_table: bool) -> None:
        """
        This method will insert the data contained in a list of dictionaries into a table. The query itself is handled
        by SQLAlchemy, so it's only needed to pass the correct data type.
        """
        cls.__delete_all_rows(table, connection, delete_table)
        connection.execute(table.insert(), data)

    @classmethod
    def _insert_dataframe(
        cls, df: Union[pd.DataFrame, pd.Series], table: Any, connection: Any, delete_table: bool
    ) -> None:
        if isinstance(df, pd.Series):
            data = [df.to_dict()]
        elif isinstance(df, pd.DataFrame):
            data = df.to_dict(orient="records")
        cls._insert_dicts(data, table, connection, delete_table)

    @classmethod
    def __delete_all_rows(cls, table: Any, connection: Any, delete_table: bool) -> None:
        if delete_table:
            connection.execute(table.delete())



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any, Optional


class Reason:
    """A reason explains why a specific action cannot be performed.

    This is a parent class aiming at being implemented by specific subclasses.

    Because Taipy applications are natively multiuser, asynchronous, and dynamic,
    some functions might not be called in some specific contexts. You can protect
    such calls by calling other methods that return a Reasons object. It acts like a
    boolean: True if the operation can be performed and False otherwise.
    If the action cannot be performed, the Reasons object holds all the `reasons as a list
    of `Reason` objects. Each `Reason` holds an explanation of why the operation cannot be
    performed.

    Attributes:
        reason (str): The English representation of the reason why the action cannot be performed.
    """

    def __init__(self, reason: str):
        self._reason = reason

    def __str__(self) -> str:
        return self._reason

    def __repr__(self) -> str:
        return self._reason

    def __hash__(self) -> int:
        return hash(self._reason)

    def __eq__(self, value: Any) -> bool:
        return isinstance(value, Reason) and value._reason == self._reason


class _DataNodeReasonMixin:
    def __init__(self, datanode_id: str):
        self.datanode_id = datanode_id

    @property
    def datanode(self):
        from ..data._data_manager_factory import _DataManagerFactory

        return _DataManagerFactory._build_manager()._get(self.datanode_id)


class DataNodeEditInProgress(Reason, _DataNodeReasonMixin):
    """
    A `DataNode^` is being edited, which prevents specific actions from being performed.

    Attributes:
        datanode_id (str): The identifier of the `DataNode^`.
    """

    def __init__(self, datanode_id: str):
        Reason.__init__(self, f"DataNode {datanode_id} is being edited")
        _DataNodeReasonMixin.__init__(self, datanode_id)


class DataNodeIsNotWritten(Reason, _DataNodeReasonMixin):
    """
    A `DataNode^` has never been written, which prevents specific actions from being performed.

    Attributes:
        datanode_id (str): The identifier of the `DataNode^`.
    """

    def __init__(self, datanode_id: str):
        Reason.__init__(self, f"DataNode {datanode_id} is not written")
        _DataNodeReasonMixin.__init__(self, datanode_id)


class EntityIsNotSubmittableEntity(Reason):
    """
    An entity is not a submittable entity, which prevents specific actions from being performed.

    Attributes:
        entity_id (str): The entity identifier.
    """

    def __init__(self, entity_id: str):
        Reason.__init__(self, f"Entity {entity_id} is not a submittable entity")


class WrongConfigType(Reason):
    """
    A config id is not a valid expected config, which prevents specific actions from being performed.

    Attributes:
        config_id (str): The identifier of the config.
        config_type (str): The expected config type.
    """

    def __init__(self, config_id: str, config_type: Optional[str]):
        if config_type:
            reason = f'Object "{config_id}" must be a valid {config_type}'
        else:
            reason = f'Object "{config_id}" is not a valid config to be created'

        Reason.__init__(self, reason)


class NotGlobalScope(Reason):
    """
    A data node config does not have a GLOBAL scope, which prevents specific actions from being performed.

    Attributes:
        config_id (str): The identifier of the config.
    """

    def __init__(self, config_id: str):
        Reason.__init__(self, f'Data node config "{config_id}" does not have GLOBAL scope')


class UploadFileCanNotBeRead(Reason, _DataNodeReasonMixin):
    """
    The uploaded file can not be read, therefore is not a valid data file for the data node.

    Attributes:
        file_name (str): The name of the file that was uploaded.
        datanode_id (str): The datanode id that the file is intended to upload to.
    """

    def __init__(self, file_name: str, datanode_id: str):
        Reason.__init__(
            self,
            f"The uploaded file {file_name} can not be read, "
            f'therefore is not a valid data file for data node "{datanode_id}"',
        )
        _DataNodeReasonMixin.__init__(self, datanode_id)


class NoFileToDownload(Reason, _DataNodeReasonMixin):
    """
    There is no file to download, therefore the download action cannot be performed.

    Attributes:
        datanode_id (str): The id of the data node that the file is intended to download from.
    """

    def __init__(self, file_path: str, datanode_id: str):
        Reason.__init__(
            self, f"Path '{file_path}' from data node '{datanode_id}'" f" does not exist and cannot be downloaded"
        )
        _DataNodeReasonMixin.__init__(self, datanode_id)


class NotAFile(Reason, _DataNodeReasonMixin):
    """
    The data node path is not a file, therefore the download action cannot be performed.

    Attributes:
        datanode_id (str): The datanode id that the file is intended to download from.
    """

    def __init__(self, file_path: str, datanode_id: str):
        Reason.__init__(
            self, f"Path '{file_path}' from data node '{datanode_id}'" f" is not a file and can t be downloaded"
        )
        _DataNodeReasonMixin.__init__(self, datanode_id)


class InvalidUploadFile(Reason, _DataNodeReasonMixin):
    """
    The uploaded file has invalid data, therefore is not a valid data file for the data node.

    Attributes:
        file_name (str): The name of the file that was uploaded.
        datanode_id (str): The datanode id that the file is intended to upload to.
    """

    def __init__(self, file_name: str, datanode_id: str):
        Reason.__init__(self, f'The uploaded file {file_name} has invalid data for data node "{datanode_id}"')
        _DataNodeReasonMixin.__init__(self, datanode_id)


class EntityDoesNotExist(Reason, _DataNodeReasonMixin):
    """
    The entity id provided does not exist in the repository.

    Attributes:
        entity_id (str): The entity identifier.
    """

    def __init__(self, entity_id: str):
        Reason.__init__(self, f"Entity {entity_id} does not exist in the repository")


class JobIsNotFinished(Reason, _DataNodeReasonMixin):
    """
    As the job is not finished yet, it prevents specific actions from being performed.

    Attributes:
        job_id (str): The job identifier.
    """

    def __init__(self, job_id: str):
        Reason.__init__(self, f"The job {job_id} is not finished yet")


class EntityIsNotAScenario(Reason, _DataNodeReasonMixin):
    """
    The entity is not a scenario, which prevents specific actions from being performed.

    Attributes:
        entity_id (str): The entity identifier.
    """

    def __init__(self, entity_id: str):
        Reason.__init__(self, f"The entity {entity_id} is not a scenario")


class ScenarioIsThePrimaryScenario(Reason, _DataNodeReasonMixin):
    """
    The scenario is the primary scenario of a cycle, which prevents specific actions from being performed.

    Attributes:
        scenario_id (str): The scenario identifier.
        cycle_id (str): The cycle identifier.
    """

    def __init__(self, scenario_id: str, cycle: str):
        Reason.__init__(self, f"The scenario {scenario_id} is the primary scenario of cycle {cycle}")


class ScenarioDoesNotBelongToACycle(Reason, _DataNodeReasonMixin):
    """
    The scenario does not belong to any cycle, which prevents specific actions from being performed.

    Attributes:
        scenario_id (str): The scenario identifier.
    """

    def __init__(self, scenario_id: str):
        Reason.__init__(self, f"The scenario {scenario_id} does not belong to any cycle")


class SubmissionIsNotFinished(Reason, _DataNodeReasonMixin):
    """
    The submission is not finished yet.

    Attributes:
        submission_id (str): The submission identifier.
    """

    def __init__(self, submission_id: str):
        Reason.__init__(self, f"The submission {submission_id} is not finished yet")


class SubmissionStatusIsUndefined(Reason, _DataNodeReasonMixin):
    """
    The submission status is undefined.

    Attributes:
        submission_id (str): The submission identifier.
    """

    def __init__(self, submission_id: str):
        Reason.__init__(self, f"The status of submission {submission_id} is undefined")



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from collections.abc import Hashable
from functools import reduce
from itertools import chain
from operator import and_, or_
from typing import Dict, Iterable, List, Tuple, Union

import numpy as np
import pandas as pd
from pandas.core.common import is_bool_indexer

from .operator import JoinOperator, Operator


class _FilterDataNode:
    @staticmethod
    def __is_pandas_object(data) -> bool:
        return isinstance(data, (pd.DataFrame, pd.Series))

    @staticmethod
    def __is_multi_sheet_excel(data) -> bool:
        if isinstance(data, Dict):
            has_df_children = all(isinstance(e, pd.DataFrame) for e in data.values())
            has_list_children = all(isinstance(e, List) for e in data.values())
            has_np_array_children = all(isinstance(e, np.ndarray) for e in data.values())
            return has_df_children or has_list_children or has_np_array_children
        return False

    @staticmethod
    def __is_list_of_dict(data) -> bool:
        return all(isinstance(x, Dict) for x in data)

    @staticmethod
    def _filter_by_key(data, key):
        if isinstance(key, int):
            return _FilterDataNode.__getitem_int(data, key)

        if isinstance(key, slice) or (isinstance(key, tuple) and any(isinstance(e, slice) for e in key)):
            return _FilterDataNode.__getitem_slice(data, key)

        if isinstance(key, Hashable):
            return _FilterDataNode.__getitem_hashable(data, key)

        if isinstance(key, pd.DataFrame):
            return _FilterDataNode.__getitem_dataframe(data, key)

        if is_bool_indexer(key):
            return _FilterDataNode.__getitem_bool_indexer(data, key)

        if isinstance(key, Iterable):
            return _FilterDataNode.__getitem_iterable(data, key)

        return None

    @staticmethod
    def __getitem_int(data, key):
        return data[key]

    @staticmethod
    def __getitem_hashable(data, key):
        if _FilterDataNode.__is_pandas_object(data) or _FilterDataNode.__is_multi_sheet_excel(data):
            return data.get(key)
        return [getattr(entry, key, None) for entry in data]

    @staticmethod
    def __getitem_slice(data, key):
        return data[key]

    @staticmethod
    def __getitem_dataframe(data, key: pd.DataFrame):
        if _FilterDataNode.__is_pandas_object(data):
            return data[key]
        if _FilterDataNode.__is_list_of_dict(data):
            filtered_data = []
            for i, row in key.iterrows():
                filtered_row = {}
                for col in row.index:
                    filtered_row[col] = data[i][col] if row[col] else None
                filtered_data.append(filtered_row)
            return filtered_data
        return None

    @staticmethod
    def __getitem_bool_indexer(data, key):
        if _FilterDataNode.__is_pandas_object(data):
            return data[key]
        return [e for i, e in enumerate(data) if key[i]]

    @staticmethod
    def __getitem_iterable(data, keys):
        if _FilterDataNode.__is_pandas_object(data):
            return data[keys]

        return [{k: getattr(entry, k) for k in keys if hasattr(entry, k)} for entry in data]

    @staticmethod
    def _filter(data, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        if len(operators) == 0:
            return data

        if isinstance(data, Dict):
            return {k: _FilterDataNode._filter(v, operators, join_operator) for k, v in data.items()}

        if not isinstance(operators[0], (list, tuple)):
            if isinstance(data, pd.DataFrame):
                return _FilterDataNode.__filter_dataframe_per_key_value(data, operators[0], operators[1], operators[2])
            if isinstance(data, np.ndarray):
                list_operators = [operators]
                return _FilterDataNode.__filter_numpy_array(data, list_operators)
            if isinstance(data, List):
                return _FilterDataNode.__filter_list_per_key_value(data, operators[0], operators[1], operators[2])
        else:
            if isinstance(data, pd.DataFrame):
                return _FilterDataNode.__filter_dataframe(data, operators, join_operator=join_operator)
            if isinstance(data, np.ndarray):
                return _FilterDataNode.__filter_numpy_array(data, operators, join_operator=join_operator)
            if isinstance(data, List):
                return _FilterDataNode.__filter_list(data, operators, join_operator=join_operator)
        raise NotImplementedError

    @staticmethod
    def __filter_dataframe(df_data: pd.DataFrame, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        if join_operator == JoinOperator.AND:
            how = "inner"
        elif join_operator == JoinOperator.OR:
            how = "outer"
        else:
            raise NotImplementedError

        filtered_df_data = [
            _FilterDataNode.__filter_dataframe_per_key_value(df_data, key, value, operator)
            for key, value, operator in operators
        ]

        return _FilterDataNode.__dataframe_merge(filtered_df_data, how) if filtered_df_data else pd.DataFrame()

    @staticmethod
    def __filter_dataframe_per_key_value(df_data: pd.DataFrame, key: str, value, operator: Operator):
        df_by_col = df_data[key]
        if operator == Operator.EQUAL:
            df_by_col = df_by_col == value
        if operator == Operator.NOT_EQUAL:
            df_by_col = df_by_col != value
        if operator == Operator.LESS_THAN:
            df_by_col = df_by_col < value
        if operator == Operator.LESS_OR_EQUAL:
            df_by_col = df_by_col <= value
        if operator == Operator.GREATER_THAN:
            df_by_col = df_by_col > value
        if operator == Operator.GREATER_OR_EQUAL:
            df_by_col = df_by_col >= value
        return df_data[df_by_col]

    @staticmethod
    def __dataframe_merge(df_list: List, how="inner"):
        return reduce(lambda df1, df2: pd.merge(df1, df2, how=how), df_list)

    @staticmethod
    def __filter_numpy_array(data: np.ndarray, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        conditions = [
            _FilterDataNode.__get_filter_condition_per_key_value(data, key, value, operator)
            for key, value, operator in operators
        ]

        if join_operator == JoinOperator.AND:
            join_conditions = reduce(and_, conditions)
        elif join_operator == JoinOperator.OR:
            join_conditions = reduce(or_, conditions)
        else:
            raise NotImplementedError

        return data[join_conditions]

    @staticmethod
    def __get_filter_condition_per_key_value(array_data: np.ndarray, key, value, operator: Operator):
        if not isinstance(key, int):
            key = int(key)

        if operator == Operator.EQUAL:
            return array_data[:, key] == value
        if operator == Operator.NOT_EQUAL:
            return array_data[:, key] != value
        if operator == Operator.LESS_THAN:
            return array_data[:, key] < value
        if operator == Operator.LESS_OR_EQUAL:
            return array_data[:, key] <= value
        if operator == Operator.GREATER_THAN:
            return array_data[:, key] > value
        if operator == Operator.GREATER_OR_EQUAL:
            return array_data[:, key] >= value

        raise NotImplementedError

    @staticmethod
    def __filter_list(list_data: List, operators: Union[List, Tuple], join_operator=JoinOperator.AND):
        filtered_list_data = [
            _FilterDataNode.__filter_list_per_key_value(list_data, key, value, operator)
            for key, value, operator in operators
        ]
        if not filtered_list_data:
            return filtered_list_data

        if join_operator == JoinOperator.AND:
            return _FilterDataNode.__list_intersect(filtered_list_data)
        elif join_operator == JoinOperator.OR:
            merged_list = list(chain.from_iterable(filtered_list_data))
            if all(isinstance(e, Dict) for e in merged_list):
                return list({frozenset(item.items()) for item in merged_list})
            return list(set(merged_list))
        else:
            raise NotImplementedError

    @staticmethod
    def __filter_list_per_key_value(list_data: List, key: str, value, operator: Operator):
        filtered_list = []
        for row in list_data:
            if isinstance(row, Dict):
                row_value = row.get(key, None)
            else:
                row_value = getattr(row, key, None)

            if operator == Operator.EQUAL and row_value == value:
                filtered_list.append(row)
            if operator == Operator.NOT_EQUAL and row_value != value:
                filtered_list.append(row)
            if operator == Operator.LESS_THAN and row_value < value:
                filtered_list.append(row)
            if operator == Operator.LESS_OR_EQUAL and row_value <= value:
                filtered_list.append(row)
            if operator == Operator.GREATER_THAN and row_value > value:
                filtered_list.append(row)
            if operator == Operator.GREATER_OR_EQUAL and row_value >= value:
                filtered_list.append(row)
        return filtered_list

    @staticmethod
    def __list_intersect(list_data):
        return list(set(list_data.pop()).intersection(*map(set, list_data)))



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
from typing import Dict, Iterable, List, Optional, Set, Union

from taipy.common.config import Config
from taipy.common.config._config import _Config
from taipy.common.config.common.scope import Scope

from .._manager._manager import _Manager
from .._version._version_mixin import _VersionMixin
from ..config.data_node_config import DataNodeConfig
from ..cycle.cycle_id import CycleId
from ..exceptions.exceptions import InvalidDataNodeType
from ..notification import Event, EventEntityType, EventOperation, Notifier, _make_event
from ..reason import NotGlobalScope, ReasonCollection, WrongConfigType
from ..scenario.scenario_id import ScenarioId
from ..sequence.sequence_id import SequenceId
from ._data_fs_repository import _DataFSRepository
from ._file_datanode_mixin import _FileDataNodeMixin
from .data_node import DataNode
from .data_node_id import DataNodeId


class _DataManager(_Manager[DataNode], _VersionMixin):
    _DATA_NODE_CLASS_MAP = DataNode._class_map()  # type: ignore
    _ENTITY_NAME = DataNode.__name__
    _EVENT_ENTITY_TYPE = EventEntityType.DATA_NODE
    _repository: _DataFSRepository

    @classmethod
    def _bulk_get_or_create(
        cls,
        data_node_configs: List[DataNodeConfig],
        cycle_id: Optional[CycleId] = None,
        scenario_id: Optional[ScenarioId] = None,
    ) -> Dict[DataNodeConfig, DataNode]:
        data_node_configs = [Config.data_nodes[dnc.id] for dnc in data_node_configs]
        dn_configs_and_owner_id = []
        for dn_config in data_node_configs:
            scope = dn_config.scope
            owner_id: Union[Optional[SequenceId], Optional[ScenarioId], Optional[CycleId]]
            if scope == Scope.SCENARIO:
                owner_id = scenario_id
            elif scope == Scope.CYCLE:
                owner_id = cycle_id
            else:
                owner_id = None
            dn_configs_and_owner_id.append((dn_config, owner_id))

        data_nodes = cls._repository._get_by_configs_and_owner_ids(
            dn_configs_and_owner_id, cls._build_filters_with_version(None)
        )

        return {
            dn_config: data_nodes.get((dn_config, owner_id)) or cls._create_and_set(dn_config, owner_id, None)
            for dn_config, owner_id in dn_configs_and_owner_id
        }

    @classmethod
    def _can_create(cls, config: Optional[DataNodeConfig] = None) -> ReasonCollection:
        config_id = getattr(config, "id", None) or str(config)
        reason_collection = ReasonCollection()

        if config is not None:
            if not isinstance(config, DataNodeConfig):
                reason_collection._add_reason(config_id, WrongConfigType(config_id, DataNodeConfig.__name__))
            elif config.scope is not Scope.GLOBAL:
                reason_collection._add_reason(config_id, NotGlobalScope(config_id))

        return reason_collection

    @classmethod
    def _create_and_set(
        cls, data_node_config: DataNodeConfig, owner_id: Optional[str], parent_ids: Optional[Set[str]]
    ) -> DataNode:
        data_node = cls.__create(data_node_config, owner_id, parent_ids)
        cls._set(data_node)
        Notifier.publish(_make_event(data_node, EventOperation.CREATION))
        return data_node

    @classmethod
    def __create(
        cls, data_node_config: DataNodeConfig, owner_id: Optional[str], parent_ids: Optional[Set[str]]
    ) -> DataNode:
        try:
            version = cls._get_latest_version()
            props = data_node_config._properties.copy()

            if data_node_config.storage_type:
                storage_type = data_node_config.storage_type
            else:
                storage_type = Config.sections[DataNodeConfig.name][_Config.DEFAULT_KEY].storage_type

            return cls._DATA_NODE_CLASS_MAP[storage_type](
                config_id=data_node_config.id,
                scope=data_node_config.scope or DataNodeConfig._DEFAULT_SCOPE,
                validity_period=data_node_config.validity_period,
                owner_id=owner_id,
                parent_ids=parent_ids,
                version=version,
                properties=props,
            )
        except KeyError:
            raise InvalidDataNodeType(data_node_config.storage_type) from None

    @classmethod
    def _get_all(cls, version_number: Optional[str] = None) -> List[DataNode]:
        """
        Returns all entities.
        """
        filters = cls._build_filters_with_version(version_number)
        return cls._repository._load_all(filters)

    @classmethod
    def _clean_generated_file(cls, data_node: DataNode) -> None:
        if not isinstance(data_node, _FileDataNodeMixin):
            return
        if data_node.is_generated and os.path.exists(data_node.path):
            os.remove(data_node.path)

    @classmethod
    def _clean_generated_files(cls, data_nodes: Iterable[DataNode]) -> None:
        for data_node in data_nodes:
            cls._clean_generated_file(data_node)

    @classmethod
    def _delete(cls, data_node_id: DataNodeId) -> None:
        if data_node := cls._get(data_node_id, None):
            cls._clean_generated_file(data_node)
        super()._delete(data_node_id)

    @classmethod
    def _delete_many(cls, data_node_ids: Iterable[DataNodeId]) -> None:
        data_nodes = []
        for data_node_id in data_node_ids:
            if data_node := cls._get(data_node_id):
                data_nodes.append(data_node)
        cls._clean_generated_files(data_nodes)
        super()._delete_many(data_node_ids)

    @classmethod
    def _delete_all(cls) -> None:
        data_nodes = cls._get_all()
        cls._clean_generated_files(data_nodes)
        super()._delete_all()

    @classmethod
    def _delete_by_version(cls, version_number: str) -> None:
        data_nodes = cls._get_all(version_number)
        cls._clean_generated_files(data_nodes)
        cls._repository._delete_by(attribute="version", value=version_number)
        Notifier.publish(
            Event(EventEntityType.DATA_NODE, EventOperation.DELETION, metadata={"delete_by_version": version_number})
        )

    @classmethod
    def _get_by_config_id(cls, config_id: str, version_number: Optional[str] = None) -> List[DataNode]:
        """
        Get all datanodes by its config id.
        """
        filters = cls._build_filters_with_version(version_number)
        if not filters:
            filters = [{}]
        for fil in filters:
            fil.update({"config_id": config_id})
        return cls._repository._load_all(filters)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from functools import lru_cache

import pymongo


@lru_cache
def _connect_mongodb(
    db_host: str, db_port: int, db_username: str, db_password: str, db_extra_args: frozenset, db_driver: str
) -> pymongo.MongoClient:
    """Create a connection to a Mongo database.
    The `"mongodb_extra_args"` passed by the user is originally a dictionary, but since `@lru_cache` wrapper only
    accepts hashable parameters, the `"mongodb_extra_args"` should be converted into a frozenset beforehand.

    Parameters:
        db_host (str): the database host.
        db_port (int): the database port.
        db_username (str): the database username.
        db_password (str): the database password.
        db_extra_args (frozenset): A frozenset converted from a dictionary of additional arguments to be passed into
            database connection string.

    Returns:
        pymongo.MongoClient
    """
    auth_str = ""
    if db_username and db_password:
        auth_str = f"{db_username}:{db_password}@"

    extra_args_str = "&".join(f"{k}={str(v)}" for k, v in db_extra_args)
    if extra_args_str:
        extra_args_str = "/?" + extra_args_str

    driver = "mongodb"
    if db_driver:
        driver = f"{driver}+{db_driver}"

    connection_string = f"{driver}://{auth_str}{db_host}"
    connection_string = connection_string if db_driver else f"{connection_string}:{db_port}"
    connection_string += extra_args_str

    return pymongo.MongoClient(connection_string)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Dict, Generic, Iterable, List, Optional, TypeVar, Union

from taipy.common.logger._taipy_logger import _TaipyLogger

from .._entity._entity_ids import _EntityIds
from .._repository._abstract_repository import _AbstractRepository
from ..exceptions.exceptions import ModelNotFound
from ..notification import Event, EventOperation, Notifier
from ..reason import EntityDoesNotExist, ReasonCollection

EntityType = TypeVar("EntityType")


class _Manager(Generic[EntityType]):
    _repository: _AbstractRepository
    _logger = _TaipyLogger._get_logger()
    _ENTITY_NAME: str = "Entity"

    @classmethod
    def _delete_all(cls):
        """
        Deletes all entities.
        """
        cls._repository._delete_all()
        if hasattr(cls, "_EVENT_ENTITY_TYPE"):
            Notifier.publish(
                Event(
                    cls._EVENT_ENTITY_TYPE,
                    EventOperation.DELETION,
                    metadata={"delete_all": True},
                )
            )

    @classmethod
    def _delete_many(cls, ids: Iterable):
        """
        Deletes entities by a list of ids.
        """
        cls._repository._delete_many(ids)
        if hasattr(cls, "_EVENT_ENTITY_TYPE"):
            for entity_id in ids:
                Notifier.publish(
                    Event(
                        cls._EVENT_ENTITY_TYPE,  # type: ignore
                        EventOperation.DELETION,
                        entity_id=entity_id,
                        metadata={"delete_all": True},
                    )
                )

    @classmethod
    def _delete_by_version(cls, version_number: str):
        """
        Deletes entities by version number.
        """
        cls._repository._delete_by(attribute="version", value=version_number)
        if hasattr(cls, "_EVENT_ENTITY_TYPE"):
            Notifier.publish(
                Event(
                    cls._EVENT_ENTITY_TYPE,  # type: ignore
                    EventOperation.DELETION,
                    metadata={"delete_by_version": version_number},
                )
            )

    @classmethod
    def _delete(cls, id):
        """
        Deletes an entity by id.
        """
        cls._repository._delete(id)
        if hasattr(cls, "_EVENT_ENTITY_TYPE"):
            Notifier.publish(
                Event(
                    cls._EVENT_ENTITY_TYPE,
                    EventOperation.DELETION,
                    entity_id=id,
                )
            )

    @classmethod
    def _set(cls, entity: EntityType):
        """
        Save or update an entity.
        """
        cls._repository._save(entity)

    @classmethod
    def _get_all(cls, version_number: Optional[str] = "all") -> List[EntityType]:
        """
        Returns all entities.
        """
        filters: List[Dict] = []
        return cls._repository._load_all(filters)

    @classmethod
    def _get_all_by(cls, filters: Optional[List[Dict]] = None) -> List[EntityType]:
        """
        Returns all entities based on a criteria.
        """
        if not filters:
            filters = []
        return cls._repository._load_all(filters)

    @classmethod
    def _get(cls, entity: Union[str, EntityType], default=None) -> EntityType:
        """
        Returns an entity by id or reference.
        """
        entity_id = entity if isinstance(entity, str) else entity.id  # type: ignore
        try:
            return cls._repository._load(entity_id)
        except ModelNotFound:
            cls._logger.error(f"{cls._ENTITY_NAME} not found: {entity_id}")
            return default

    @classmethod
    def _exists(cls, entity_id: str) -> ReasonCollection:
        """
        A ReasonCollection object that can function as a Boolean value,
        which is True if the entity id exists.
        """
        reason_collector = ReasonCollection()

        if not cls._repository._exists(entity_id):
            reason_collector._add_reason(entity_id, EntityDoesNotExist(entity_id))

        return reason_collector

    @classmethod
    def _delete_entities_of_multiple_types(cls, _entity_ids: _EntityIds):
        """
        Deletes entities of multiple types.
        """
        from ..cycle._cycle_manager_factory import _CycleManagerFactory
        from ..data._data_manager_factory import _DataManagerFactory
        from ..job._job_manager_factory import _JobManagerFactory
        from ..scenario._scenario_manager_factory import _ScenarioManagerFactory
        from ..sequence._sequence_manager_factory import _SequenceManagerFactory
        from ..submission._submission_manager_factory import _SubmissionManagerFactory
        from ..task._task_manager_factory import _TaskManagerFactory

        _CycleManagerFactory._build_manager()._delete_many(_entity_ids.cycle_ids)
        _SequenceManagerFactory._build_manager()._delete_many(_entity_ids.sequence_ids)
        _ScenarioManagerFactory._build_manager()._delete_many(_entity_ids.scenario_ids)
        _TaskManagerFactory._build_manager()._delete_many(_entity_ids.task_ids)
        _JobManagerFactory._build_manager()._delete_many(_entity_ids.job_ids)
        _DataManagerFactory._build_manager()._delete_many(_entity_ids.data_node_ids)
        _SubmissionManagerFactory._build_manager()._delete_many(_entity_ids.submission_ids)

    @classmethod
    def _is_editable(cls, entity: Union[EntityType, str]) -> ReasonCollection:
        reason_collection = ReasonCollection()
        if cls._get(entity) is None:
            reason_collection._add_reason(str(entity), EntityDoesNotExist(str(entity)))
        return reason_collection

    @classmethod
    def _is_readable(cls, entity: Union[EntityType, str]) -> ReasonCollection:
        reason_collection = ReasonCollection()
        if cls._get(entity) is None:
            reason_collection._add_reason(str(entity), EntityDoesNotExist(str(entity)))
        return reason_collection



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from queue import SimpleQueue
from typing import Any, Dict, Optional, Set, Tuple

from ._registration import _Registration
from ._topic import _Topic
from .event import Event, EventEntityType, EventOperation


def _publish_event(
    entity_type: EventEntityType,
    operation: EventOperation,
    /,
    entity_id: Optional[str] = None,
    attribute_name: Optional[str] = None,
    attribute_value: Optional[Any] = None,
    **kwargs,
) -> None:
    """Internal helper function to send events.

    It basically creates an event corresponding to the given arguments
    and send it using `Notifier.publish(event)`

    Parameters:
        entity_type (EventEntityType^)
        operation (EventOperation^)
        entity_id (Optional[str])
        attribute_name (Optional[str])
        attribute_value (Optional[Any])
        **kwargs
    """
    event = Event(
        entity_id=entity_id,
        entity_type=entity_type,
        operation=operation,
        attribute_name=attribute_name,
        attribute_value=attribute_value,
        metadata=kwargs,
    )
    Notifier.publish(event)


class Notifier:
    """A class for managing event registrations and publishing a Taipy application events."""

    _topics_registrations_list: Dict[_Topic, Set[_Registration]] = {}

    @classmethod
    def register(
        cls,
        entity_type: Optional[EventEntityType] = None,
        entity_id: Optional[str] = None,
        operation: Optional[EventOperation] = None,
        attribute_name: Optional[str] = None,
    ) -> Tuple[str, SimpleQueue]:
        """Register a listener for a specific event topic.

        The topic is defined by the combination of an optional entity type, an optional
        entity id, an optional operation, and an optional attribute name. The purpose is
        to be as flexible as possible. For example, we can register to:

        - All scenario creations
        - A specific data node update
        - A sequence submission
        - A Scenario deletion
        - Job failures

        !!! example "Standard usage"

            ```python
            registration_id, registered_queue = Notifier.register(
                entity_type=EventEntityType.SCENARIO,
                operation=EventOperation.CREATION
            )
            ```

        Parameters:
            entity_type (Optional[EventEntityType^]): If provided, the listener will
                be notified for all events related to this entity type. Otherwise,
                the listener will be notified for events related to all entity types.
                <br>
                The possible entity type values are defined in the `EventEntityType^`
                enum. The possible values are:
                <ul>
                    <li>CYCLE</li>
                    <li>SCENARIO</li>
                    <li>SEQUENCE</li>
                    <li>TASK</li>
                    <li>DATA_NODE</li>
                    <li>JOB</li>
                    <li>SUBMISSION</li>
                </ul>
            entity_id (Optional[str]): If provided, the listener will be notified
                for all events related to this entity. Otherwise, the listener
                will be notified for events related to all entities.
            operation (Optional[EventOperation^]): If provided, the listener will
                be notified for all events related to this operation. Otherwise,
                the listener will be notified for events related to all operations.
                <br>
                The possible operation values are defined in the `EventOperation^`
                enum. The possible values are:
                <ul>
                    <li>CREATION</li>
                    <li>UPDATE</li>
                    <li>DELETION</li>
                    <li>SUBMISSION</li>
                </ul>
            attribute_name (Optional[str]): If provided, the listener will be notified
                for all events related to this entity's attribute. Otherwise, the listener
                will be notified for events related to all attributes.

        Returns:
            A tuple containing the registration id and the event queue.
        """
        registration = _Registration.from_topic(entity_type, entity_id, operation, attribute_name)
        return cls.__do_register(registration)

    @classmethod
    def unregister(cls, registration_id: str) -> None:
        """Unregister a listener.

        !!! example "Standard usage"

            ```python
            registration_id, registered_queue = Notifier.register(
                entity_type=EventEntityType.CYCLE,
                entity_id="CYCLE_cycle_1",
                operation=EventOperation.CREATION
            )

            Notifier.unregister(registration_id)
            ```

        Parameters:
            registration_id (`RegistrationId`): The registration id returned by the `register` method.
        """
        to_remove_registration: Optional[_Registration] = None

        for _, registrations in cls._topics_registrations_list.items():
            for reg in registrations:
                if reg.registration_id == registration_id:
                    to_remove_registration = reg
                    break
            if to_remove_registration:
                break

        if to_remove_registration:
            for topic in to_remove_registration.topics:
                if topic in cls._topics_registrations_list:
                    registrations = cls._topics_registrations_list[topic]
                    registrations.remove(to_remove_registration)
                    if len(registrations) == 0:
                        del cls._topics_registrations_list[topic]

    @classmethod
    def publish(cls, event: Event) -> None:
        """Publish a Taipy application event to all registered listeners whose topic matches the event.

        Parameters:
            event (`Event^`): The event to publish.
        """
        processed_registrations: Set[_Registration] = set()
        for topic, registrations in cls._topics_registrations_list.items():
            if Notifier._is_matching(event, topic):
                for registration in registrations:
                    if registration not in processed_registrations:
                        registration.queue.put(event)
                        processed_registrations.add(registration)

    @classmethod
    def _register_from_registration(cls, registration: _Registration) -> Tuple[str, SimpleQueue]:
        """Register a listener from a registration object.

        Arguments:
            registration (`_Registration`): The registration object.

        Returns:
            A tuple containing the registration id and the event queue.
        """
        return cls.__do_register(registration)

    @staticmethod
    def _is_matching(event: Event, topic: _Topic) -> bool:
        """Check if an event matches a topic."""
        if topic.entity_type is not None and event.entity_type != topic.entity_type:
            return False
        if topic.entity_id is not None and event.entity_id != topic.entity_id:
            return False
        if topic.operation is not None and event.operation != topic.operation:
            return False
        if topic.attribute_name is not None and event.attribute_name and event.attribute_name != topic.attribute_name:
            return False
        return True

    @classmethod
    def __do_register(cls, registration: _Registration) -> Tuple[str, SimpleQueue]:
        for topic in registration.topics:
            if registrations := cls._topics_registrations_list.get(topic, None):
                registrations.add(registration)
            else:
                cls._topics_registrations_list[topic] = {registration}
        return registration.registration_id, registration.queue



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import uuid
from typing import List, Optional, Union

from taipy.common.config import Config
from taipy.common.config._config_comparator._comparator_result import _ComparatorResult
from taipy.common.config.exceptions.exceptions import InconsistentEnvVariableError
from taipy.common.logger._taipy_logger import _TaipyLogger

from .._manager._manager import _Manager
from ..exceptions.exceptions import (
    ConfigCoreVersionMismatched,
    ConflictedConfigurationError,
    ModelNotFound,
    NonExistingVersion,
    VersionAlreadyExistsAsDevelopment,
)
from ..reason import ReasonCollection
from ._version import _Version
from ._version_fs_repository import _VersionFSRepository


class _VersionManager(_Manager[_Version]):
    _ENTITY_NAME = _Version.__name__

    _logger = _TaipyLogger._get_logger()

    _DEVELOPMENT_VERSION = ["development", "dev"]
    _LATEST_VERSION = "latest"
    _ALL_VERSION = ["all", ""]

    _DEFAULT_VERSION = _LATEST_VERSION

    _repository: _VersionFSRepository

    @classmethod
    def _get(cls, entity: Union[str, _Version], default=None) -> _Version:
        """
        Returns the version entity by id or reference.
        """
        entity_id = entity if isinstance(entity, str) else entity.id
        try:
            return cls._repository._load(entity_id)
        except ModelNotFound:
            return default

    @classmethod
    def _get_or_create(cls, id: str, force: bool) -> _Version:
        if version := cls._get(id):
            comparator_result = Config._comparator._find_conflict_config(version.config, Config._applied_config, id)  # type: ignore[attr-defined]
            if comparator_result.get(_ComparatorResult.CONFLICTED_SECTION_KEY):
                if not force:
                    raise ConflictedConfigurationError()

                cls._logger.warning(f"Option --force is detected, overriding the configuration of version {id} ...")
                version.config = Config._applied_config  # type: ignore[attr-defined]

        else:
            version = _Version(id=id, config=Config._applied_config)  # type: ignore[attr-defined]

        cls._set(version)
        return version

    @classmethod
    def _get_all(cls, version_number: Optional[Union[str, List]] = "all") -> List[_Version]:
        """
        Returns all entities.
        """
        version_number = cls._replace_version_number(version_number)  # type: ignore
        if not isinstance(version_number, List):
            version_number = [version_number] if version_number else []
        filters = [{"version": version} for version in version_number]
        return cls._repository._load_all(filters)

    @classmethod
    def _is_deletable(cls, version_number: Optional[str]) -> ReasonCollection:
        return ReasonCollection()

    @classmethod
    def _set_development_version(cls, version_number: str) -> str:
        cls._get_or_create(version_number, force=True)
        cls._repository._set_development_version(version_number)
        return version_number

    @classmethod
    def _get_development_version(cls) -> str:
        try:
            return cls._repository._get_development_version()
        except (FileNotFoundError, ModelNotFound):
            return cls._set_development_version(str(uuid.uuid4()))

    @classmethod
    def _set_experiment_version(cls, version_number: str, force: bool = False) -> str:
        if version_number == cls._get_development_version():
            raise VersionAlreadyExistsAsDevelopment(version_number)

        try:
            cls._get_or_create(version_number, force)
        except ConflictedConfigurationError:
            raise SystemExit(
                f"Please add a new experiment version or run your application with --force option to"
                f" override the Config of experiment {version_number}."
            ) from None
        cls._repository._set_latest_version(version_number)
        return version_number

    @classmethod
    def _get_latest_version(cls) -> str:
        try:
            return cls._repository._get_latest_version()
        except (FileNotFoundError, ModelNotFound):
            # If there is no version in the system yet, create a new version as development version
            # This set the default versioning behavior on Jupyter notebook to Development mode
            return cls._set_development_version(str(uuid.uuid4()))

    @classmethod
    def _replace_version_number(cls, version_number: Optional[str] = None):
        if version_number is None:
            return cls._replace_version_number(cls._DEFAULT_VERSION)

        if version_number == cls._LATEST_VERSION:
            return cls._get_latest_version()
        if version_number in cls._DEVELOPMENT_VERSION:
            return cls._get_development_version()
        if version_number in cls._ALL_VERSION:
            return ""

        try:
            if version := cls._get(version_number):
                return version.id
        except InconsistentEnvVariableError:  # The version exist but the Config is alternated
            return version_number
        except ConfigCoreVersionMismatched as e:
            cls._logger.error(e.message)
            raise SystemExit() from e

        raise NonExistingVersion(version_number)

    @classmethod
    def _rename_version(cls, old_version: str, new_version: str) -> None:
        version_entity = cls._get(old_version)

        if old_version == cls._get_latest_version():
            try:
                cls._set_experiment_version(new_version)
            except VersionAlreadyExistsAsDevelopment as err:
                raise SystemExit(err.message) from None
        if old_version == cls._get_development_version():
            cls._set_development_version(new_version)
        cls._delete(old_version)

        if not cls._get(new_version):
            version_entity.id = new_version
            cls._set(version_entity)

    @classmethod
    def _manage_version(cls) -> None:
        if Config.core.mode == "development":
            from ..taipy import clean_all_entities

            current_version_number = cls._get_development_version()
            cls._logger.info(f"Development mode: Clean all entities of version {current_version_number}")
            clean_all_entities(current_version_number)
            cls._set_development_version(current_version_number)

        elif Config.core.mode == "experiment":
            if Config.core.version_number:
                current_version_number = Config.core.version_number
            else:
                current_version_number = str(uuid.uuid4())

            try:
                cls._set_experiment_version(current_version_number, Config.core.force)
            except VersionAlreadyExistsAsDevelopment as err:
                raise SystemExit(err.message) from None

        else:
            raise SystemExit(f"Undefined execution mode: {Config.core.mode}.")

    @classmethod
    def _delete_entities_of_multiple_types(cls, _entity_ids):
        raise NotImplementedError



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from datetime import datetime

from .._repository._abstract_converter import _AbstractConverter
from ..job.job import Job, JobId
from ..submission._submission_model import _SubmissionModel
from ..submission.submission import Submission
from .submission import SubmissionId


class _SubmissionConverter(_AbstractConverter):
    @classmethod
    def _entity_to_model(cls, submission: Submission) -> _SubmissionModel:
        return _SubmissionModel(
            id=submission.id,
            entity_id=submission._entity_id,
            entity_type=submission.entity_type,
            entity_config_id=submission._entity_config_id,
            job_ids=[job.id if isinstance(job, Job) else JobId(str(job)) for job in list(submission._jobs)],
            properties=submission._properties.data.copy(),
            creation_date=submission._creation_date.isoformat(),
            submission_status=submission._submission_status,
            version=submission._version,
            is_abandoned=submission._is_abandoned,
            is_completed=submission._is_completed,
            is_canceled=submission._is_canceled,
            running_jobs=list(submission._running_jobs),
            blocked_jobs=list(submission._blocked_jobs),
            pending_jobs=list(submission._pending_jobs),
        )

    @classmethod
    def _model_to_entity(cls, model: _SubmissionModel) -> Submission:
        submission = Submission(
            entity_id=model.entity_id,
            entity_type=model.entity_type,
            entity_config_id=model.entity_config_id,
            id=SubmissionId(model.id),
            jobs=model.job_ids,
            properties=model.properties,
            creation_date=datetime.fromisoformat(model.creation_date),
            submission_status=model.submission_status,
            version=model.version,
        )

        submission._is_abandoned = model.is_abandoned
        submission._is_completed = model.is_completed
        submission._is_canceled = model.is_canceled

        submission._running_jobs = set(model.running_jobs)
        submission._blocked_jobs = set(model.blocked_jobs)
        submission._pending_jobs = set(model.pending_jobs)

        return submission



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""
The Taipy `core` package provides powerful, customized, data-driven back-end functionalities.
It provides the tools to help data scientists and Python developers transform their algorithms
into a complete back-end application. In  particular, it helps with:

- Data Integration
- Task Orchestration
- What-if Analysis
- Version management

More details on the Core functionalities are available in the user manual.

To use such functionalities, the first step consists of setting up the Taipy configuration to design
your application's characteristics and behaviors. Use the `Config^` singleton class (from `taipy.common.config`)
to configure your application. Please refer to the
[data nodes](../../../../userman/scenario_features/data-integration/data-node-config.md),
[tasks](../../../../userman/scenario_features/task-orchestration/scenario-config.md),
and [scenarios](../../../../userman/scenario_features/sdm/scenario/scenario-config.md) pages.

Once your application is configured, import module `import taipy as tp` so you can use any function described
in the following section on [Function](#functions). In particular, the most used functions are `tp.create_scenario()`,
`tp.get_scenarios()`, `tp.get_data_nodes()`, `tp.submit()`, used to get, create, and submit entities.

!!! Note

    Taipy provides a runnable service, `Orchestrator^` that runs as a service in a dedicated thread.
    The purpose is to have a dedicated thread responsible for dispatching the submitted jobs to an available
    executor for their execution.

    In particular, this `Orchestrator^` service is automatically run when used with Taipy REST or Taipy GUI.
    See the [running services](../../../../userman/run-deploy/run/running_services.md) page of the user
    manual for more details.
"""

from ._init import *
from ._init_version import _read_version
from .common.mongo_default_document import MongoDefaultDocument
from .data.data_node_id import Edit
from .exceptions import exceptions

__version__ = _read_version()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Dict, Set

from .reason import Reason


class ReasonCollection:
    """Class used to store all the reasons to explain why some Taipy operations are not allowed.

    Because Taipy applications are natively multiuser, asynchronous, and dynamic,
    some functions might not be called in some specific contexts. You can protect
    such calls by calling other methods that return a `ReasonCollection`. It acts like a
    boolean: True if the operation can be performed and False otherwise.
    If the action cannot be performed, the ReasonCollection holds all the individual reasons as
    a list of `Reason` objects. Each `Reason` explains why the operation cannot be performed.
    """

    def __init__(self) -> None:
        self._reasons: Dict[str, Set[Reason]] = {}

    def _add_reason(self, entity_id: str, reason: Reason) -> "ReasonCollection":
        if entity_id not in self._reasons:
            self._reasons[entity_id] = set()
        self._reasons[entity_id].add(reason)
        return self

    def _remove_reason(self, entity_id: str, reason: Reason) -> "ReasonCollection":
        if entity_id in self._reasons and reason in self._reasons[entity_id]:
            self._reasons[entity_id].remove(reason)
            if len(self._reasons[entity_id]) == 0:
                del self._reasons[entity_id]
        return self

    def _entity_id_exists_in_reason(self, entity_id: str) -> bool:
        return entity_id in self._reasons

    def __bool__(self) -> bool:
        return len(self._reasons) == 0

    @property
    def reasons(self) -> str:
        """Retrieves a collection of reasons as a string that explains why the action cannot be performed.

        Returns:
            A string that contains all the reasons why the action cannot be performed.
        """
        if self._reasons:
            return "; ".join("; ".join([str(reason) for reason in reasons]) for reasons in self._reasons.values()) + "."
        return ""



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import Any, List

from taipy.common.config import Config
from taipy.common.config._serializer._toml_serializer import _TomlSerializer
from taipy.common.logger._taipy_logger import _TaipyLogger

from ...data._data_manager_factory import _DataManagerFactory
from ...data.data_node import DataNode
from ...exceptions import DataNodeWritingError
from ...job.job_id import JobId
from ...task.task import Task

logger = _TaipyLogger._get_logger()


class _TaskFunctionWrapper:
    """Wrapper around task function."""

    def __init__(self, job_id: JobId, task: Task):
        self.job_id = job_id
        self.task = task

    def __call__(self, **kwargs):
        """Make this object callable as a function. Actually calls `execute`."""
        return self.execute(**kwargs)

    def execute(self, **kwargs):
        """Execute the wrapped function. If `config_as_string` is given, then it will be reapplied to the config."""
        try:
            if config_as_string := kwargs.pop("config_as_string", None):
                Config._applied_config._update(_TomlSerializer()._deserialize(config_as_string))
                Config.block_update()

            inputs = list(self.task.input.values())
            outputs = list(self.task.output.values())

            arguments = self._read_inputs(inputs)
            results = self._execute_fct(arguments)
            return self._write_data(outputs, results, self.job_id)
        except Exception as e:
            logger.error("Error during task function execution!", exc_info=1)
            return [e]

    def _read_inputs(self, inputs: List[DataNode]) -> List[Any]:
        data_manager = _DataManagerFactory._build_manager()
        return [data_manager._get(dn.id).read_or_raise() for dn in inputs]

    def _write_data(self, outputs: List[DataNode], results, job_id: JobId):
        data_manager = _DataManagerFactory._build_manager()
        try:
            if outputs:
                _results = self._extract_results(outputs, results)
                exceptions = []
                for res, dn in zip(_results, outputs):
                    try:
                        data_node = data_manager._get(dn.id)
                        data_node._write(res)
                    except Exception as e:
                        logger.error("Error during write", exc_info=1)
                        exceptions.append(DataNodeWritingError(f"Error writing in datanode id {dn.id}: {e}"))
                return exceptions
        except Exception as e:
            return [e]

    def _execute_fct(self, arguments: List[Any]) -> Any:
        return self.task.function(*arguments)

    def _extract_results(self, outputs: List[DataNode], results: Any) -> List[Any]:
        _results: List[Any] = [results] if len(outputs) == 1 else results
        if len(_results) != len(outputs):
            raise DataNodeWritingError("Error: wrong number of result or task output")
        return _results



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List

from .._repository._base_taipy_model import _BaseModel
from .job_id import JobId
from .status import Status


@dataclass
class _JobModel(_BaseModel):
    id: JobId
    task_id: str
    status: Status
    status_change_records: Dict[str, str]
    force: bool
    submit_id: str
    submit_entity_id: str
    creation_date: str
    subscribers: List[Dict]
    stacktrace: List[str]
    version: str

    @staticmethod
    def from_dict(data: Dict[str, Any]):
        return _JobModel(
            id=data["id"],
            task_id=data["task_id"],
            status=Status._from_repr(data["status"]),
            status_change_records=_BaseModel._deserialize_attribute(data["status_change_records"]),
            force=data["force"],
            submit_id=data["submit_id"],
            submit_entity_id=data["submit_entity_id"],
            creation_date=data["creation_date"],
            subscribers=_BaseModel._deserialize_attribute(data["subscribers"]),
            stacktrace=_BaseModel._deserialize_attribute(data["stacktrace"]),
            version=data["version"],
        )

    def to_list(self):
        return [
            self.id,
            self.task_id,
            repr(self.status),
            _BaseModel._serialize_attribute(self.status_change_records),
            self.force,
            self.submit_id,
            self.submit_entity_id,
            self.creation_date,
            _BaseModel._serialize_attribute(self.subscribers),
            _BaseModel._serialize_attribute(self.stacktrace),
            self.version,
        ]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from importlib import import_module
from operator import attrgetter
from typing import Type

from taipy.common._cli._base_cli._abstract_cli import _AbstractCLI

from ..common._check_dependencies import EnterpriseEditionUtils
from ._core_cli import _CoreCLI


class _CoreCLIFactory:
    @staticmethod
    def _build_cli() -> Type[_AbstractCLI]:
        if EnterpriseEditionUtils._using_enterprise():
            module = import_module(EnterpriseEditionUtils._TAIPY_ENTERPRISE_CORE_MODULE + "._cli._core_cli")
            core_cli = attrgetter("_CoreCLI")(module)
        else:
            core_cli = _CoreCLI

        return core_cli



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from typing import List, Optional


class ConfigCoreVersionMismatched(Exception):
    """Raised if the version in Config does not match with the version of Taipy Version management."""

    def __init__(self, config_core_version: str, core_version: str) -> None:
        self.message = (
            f"The version {config_core_version} of Taipy's entities does not match version of the Taipy"
            f" Version management {core_version}."
            f" Please update the core entities to be compatible with Taipy Core {core_version}"
            " using the `taipy migrate ...` command. For more information, please run `taipy help migrate`"
        )


class OrchestratorServiceIsAlreadyRunning(Exception):
    """Raised if the Orchestrator service is already running."""


class CycleAlreadyExists(Exception):
    """Raised if it is trying to create a Cycle that has already exists."""


class NonExistingCycle(Exception):
    """Raised if a requested cycle is not known by the Cycle manager."""

    def __init__(self, cycle_id: str):
        self.message = f"Cycle: {cycle_id} does not exist."


class MissingRequiredProperty(Exception):
    """Raised if a required property is missing when creating a Data Node."""


class InvalidDataNodeType(Exception):
    """Raised if a data node storage type does not exist."""


class MultipleDataNodeFromSameConfigWithSameOwner(Exception):
    """
    Raised if there are multiple data nodes from the same data node configuration and the same owner identifier.
    """


class NoData(Exception):
    """Raised if a data node is read before it has been written.

    This exception can be raised by `DataNode.read_or_raise()^`.
    """


class UnknownDatabaseEngine(Exception):
    """Raised if the database engine is not known when creating a connection with a SQLDataNode."""


class MissingAppendQueryBuilder(Exception):
    """Raised if no append query build is provided when appending data to a SQLDataNode."""


class UnknownParquetEngine(Exception):
    """Raised if the parquet engine is not known or not supported when create a ParquetDataNode."""


class UnknownCompressionAlgorithm(Exception):
    """Raised if the compression algorithm is not supported by ParquetDataNode."""


class NonExistingDataNode(Exception):
    """Raised if a requested DataNode is not known by the DataNode Manager."""

    def __init__(self, data_node_id: str):
        self.message = f"DataNode: {data_node_id} does not exist."


class DataNodeIsBeingEdited(Exception):
    """Raised if a DataNode is being edited."""

    def __init__(self, data_node_id: str, editor_id: Optional[str] = None):
        self.message = f"DataNode {data_node_id} is being edited{ ' by ' + editor_id if editor_id else ''}."


class NonExistingDataNodeConfig(Exception):
    """Raised if a requested DataNodeConfig is not known by the DataNode Manager."""

    def __init__(self, data_node_config_id: str):
        self.message = f"Data node config: {data_node_config_id} does not exist."


class NonExistingExcelSheet(Exception):
    """Raised if a requested Sheet name does not exist in the provided Excel file."""

    def __init__(self, sheet_name: str, excel_file_name: str):
        self.message = f"{sheet_name} does not exist in {excel_file_name}."


class ExposedTypeLengthMismatch(Exception):
    """Raised if length of exposed type list does not match with number of sheets in the provided Excel file."""


class SheetNameLengthMismatch(Exception):
    """Raised if length of sheet_name list does not match
    with number of sheets in the data to be written to Excel file."""


class InvalidExposedType(Exception):
    """Raised if an invalid exposed type is provided."""


class InvalidCustomDocument(Exception):
    """Raised if an invalid custom document class is provided to a `MongoCollectionDataNode`."""


class DataNodeConfigIsNotGlobal(Exception):
    """Raised if a DataNode is not global."""

    def __init__(self, data_node_config_id: str):
        self.message = f"Data node config `{data_node_config_id}` does not have GLOBAL scope."


class MissingReadFunction(Exception):
    """Raised if no read function is provided for the GenericDataNode."""


class MissingWriteFunction(Exception):
    """Raised if no write function is provided for the GenericDataNode."""


class JobNotDeletedException(RuntimeError):
    """Raised if there is an attempt to delete a job that cannot be deleted.

    This exception can be raised by `taipy.delete_job()^`.
    """

    def __init__(self, job_id: str):
        self.message = f"Job: {job_id} cannot be deleted."


class NonExistingJob(RuntimeError):
    """Raised if a requested job is not known by the Job manager."""

    def __init__(self, job_id: str):
        self.message = f"Job: {job_id} does not exist."


class SubmissionNotDeletedException(RuntimeError):
    """Raised if there is an attempt to delete a submission that cannot be deleted.

    This exception can be raised by `taipy.delete()^`.
    """

    def __init__(self, submission_id: str):
        self.message = f"Submission: {submission_id} cannot be deleted."


class DataNodeWritingError(RuntimeError):
    """Raised if an error happens during the writing in a data node."""


class InvalidSubscriber(RuntimeError):
    """Raised if the loaded function is not valid."""


class InvalidSequenceId(Exception):
    """Raised if a Sequence id can not be broken down."""

    def __init__(self, sequence_id: str):
        self.message = f"Sequence: {sequence_id} is invalid."


class InvalidSequence(Exception):
    """Raised if a Sequence is not a connected Directed Acyclic Graph."""

    def __init__(self, sequence_id: str):
        self.message = f"Sequence: {sequence_id} is not a connected Directed Acyclic Graph."


class NonExistingSequence(Exception):
    """Raised if a requested Sequence is not known by the Sequence Manager."""

    def __init__(self, sequence_id: str, scenario_id: str = None):
        if scenario_id:
            self.message = f"Sequence: {sequence_id} does not exist in scenario {scenario_id}."
        else:
            self.message = f"Sequence: {sequence_id} does not exist."


class SequenceAlreadyExists(Exception):
    """Raised if a Sequence already exists."""

    def __init__(self, sequence_name: str, scenario_id: str):
        self.message = f"Sequence: {sequence_name} already exists in scenario {scenario_id}."


class SequenceBelongsToNonExistingScenario(Exception):
    """Raised if a Sequence does not belong to an existing Scenario."""

    def __init__(self, sequence_id: str, scenario_id: str):
        self.message = f"Sequence: {sequence_id} belongs to a non-existing Scenario: {scenario_id}."


class SequenceTaskDoesNotExistInScenario(Exception):
    """Raised if Tasks of a Sequence do not exist in the same Scenario that the Sequence belongs to."""

    def __init__(self, task_ids: List[Optional[str]], sequence_name: str, scenario_id: str):
        self.message = f"Tasks {task_ids} of Sequence {sequence_name} does not exist in Scenario {scenario_id}."


class SequenceTaskConfigDoesNotExistInSameScenarioConfig(Exception):
    """Raised if TaskConfigs of a Sequence do not exist in the same ScenarioConfig that the Sequence belongs to."""

    def __init__(self, task_config_ids: List[Optional[str]], sequence_name: str, scenario_config_id: str):
        self.message = f"TaskConfig {task_config_ids} of Sequence name {sequence_name} "
        self.message += f"does not exist in ScenarioConfig {scenario_config_id}."


class NonExistingSequenceConfig(Exception):
    """Raised if a requested Sequence configuration is not known by the Sequence Manager."""

    def __init__(self, sequence_config_id: str):
        self.message = f"Sequence config: {sequence_config_id} does not exist."


class MultipleSequenceFromSameConfigWithSameOwner(Exception):
    """Raised if it exists multiple sequences from the same sequence config and with the same _owner_id_."""


class ModelNotFound(Exception):
    """Raised when trying to fetch a non-existent model.

    This exception can be raised by `taipy.get()^` and `taipy.delete()^`.
    """

    def __init__(self, model_name: str, model_id: str):
        self.message = f"A {model_name} model with id {model_id} could not be found."


class NonExistingScenario(Exception):
    """Raised if a requested scenario is not known by the Scenario Manager."""

    def __init__(self, scenario_id: str):
        self.message = f"Scenario: {scenario_id} does not exist."


class NonExistingScenarioConfig(Exception):
    """Raised if a requested scenario configuration is not known by the Scenario Manager.

    This exception can be raised by `taipy.compare_scenarios()^`.
    """

    def __init__(self, scenario_config_id: str):
        self.message = f"Scenario config: {scenario_config_id} does not exist."


class InvalidScenario(Exception):
    """Raised if a Scenario is not a Directed Acyclic Graph."""

    def __init__(self, scenario_id: str):
        self.message = f"Scenario: {scenario_id} is not a Directed Acyclic Graph."


class DoesNotBelongToACycle(Exception):
    """Raised if a scenario without any cycle is promoted as primary scenario."""


class DeletingPrimaryScenario(Exception):
    """Raised if a primary scenario is deleted."""


class DifferentScenarioConfigs(Exception):
    """Raised if scenario comparison is requested on scenarios with different scenario configs.

    This exception can be raised by `taipy.compare_scenarios()^`.
    """


class InsufficientScenarioToCompare(Exception):
    """Raised if too few scenarios are requested to be compared.

    Scenario comparison need at least two scenarios to compare.

    This exception can be raised by `taipy.compare_scenarios()^`.
    """


class NonExistingComparator(Exception):
    """Raised if a scenario comparator does not exist.

    This exception can be raised by `taipy.compare_scenarios()^`.
    """


class UnauthorizedTagError(Exception):
    """Must provide an authorized tag."""


class DependencyNotInstalled(Exception):
    """Raised if a package is missing."""

    def __init__(self, package_name: str):
        self.message = f"""
        Package '{package_name}' should be installed.
        Run 'pip install taipy[{package_name}]' to installed it.
        """


class NonExistingTask(Exception):
    """Raised if a requested task is not known by the Task Manager."""

    def __init__(self, task_id: str):
        self.message = f"Task: {task_id} does not exist."


class NonExistingTaskConfig(Exception):
    """Raised if a requested task configuration is not known by the Task Manager."""

    def __init__(self, id: str):
        self.message = f"Task config: {id} does not exist."


class MultipleTaskFromSameConfigWithSameOwner(Exception):
    """Raised if there are multiple tasks from the same task configuration and the same owner identifier."""


class OrchestratorNotBuilt(Exception):
    """Raised if the orchestrator was not built in the OrchestratorFactory"""


class ModeNotAvailable(Exception):
    """Raised if the mode in JobConfig is not supported."""


class NonExistingVersion(Exception):
    """Raised if request a Version that is not known by the Version Manager."""

    def __init__(self, version_number: str):
        self.message = f"Version '{version_number}' does not exist."


class VersionAlreadyExistsAsDevelopment(Exception):
    """Raised if a version already exists as the development version."""

    def __init__(self, version_number: str):
        self.message = f"Version number {version_number} already exists as the development version."
        self.message += " Please choose a different name for this experiment."


class ConflictedConfigurationError(Exception):
    """Conflicts have been detected between the current and previous Configurations."""


class InvalidEventAttributeName(Exception):
    """
    Raised if the attribute doesn't exist or an attribute name is provided
    when operation is either creation, deletion or submission
    """


class InvalidEventOperation(Exception):
    """Raised when operation doesn't belong to the entity"""


class FileCannotBeRead(Exception):
    """Raised when a file cannot be read."""


class FileEmpty(Exception):
    """Raised when a file is empty."""


class SQLQueryCannotBeExecuted(Exception):
    """Raised when an SQL Query cannot be executed."""


class _SuspiciousFileOperation(Exception):
    pass


class AttributeKeyAlreadyExisted(Exception):
    """Raised when an attribute key already existed."""

    def __init__(self, key: str):
        self.message = f"Attribute key '{key}' already existed."



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import itertools
from datetime import datetime
from queue import Queue
from threading import Lock
from time import sleep
from typing import Callable, Iterable, List, Optional, Set, Union

from taipy.common.config import Config
from taipy.common.logger._taipy_logger import _TaipyLogger

from .._entity.submittable import Submittable
from ..data._data_manager_factory import _DataManagerFactory
from ..job._job_manager_factory import _JobManagerFactory
from ..job.job import Job
from ..job.job_id import JobId
from ..submission._submission_manager_factory import _SubmissionManagerFactory
from ..submission.submission import Submission
from ..task.task import Task
from ._abstract_orchestrator import _AbstractOrchestrator


class _Orchestrator(_AbstractOrchestrator):
    """
    Handles the functional orchestrating.
    """

    jobs_to_run: Queue = Queue()
    blocked_jobs: List[Job] = []

    lock = Lock()
    __logger = _TaipyLogger._get_logger()

    @classmethod
    def initialize(cls):
        pass

    @classmethod
    def submit(
        cls,
        submittable: Submittable,
        callbacks: Optional[Iterable[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        **properties,
    ) -> Submission:
        """Submit the given `Scenario^` or `Sequence^` for an execution.

        Parameters:
             submittable (Union[Scenario^, Sequence^]): The scenario or sequence to submit for execution.
             callbacks: The optional list of functions that should be executed on jobs status change.
             force (bool) : Enforce execution of the scenario's or sequence's tasks even if their output data
                nodes are cached.
             wait (bool): Wait for the orchestrated jobs created from the scenario or sequence submission to be
                finished in asynchronous mode.
             timeout (Union[float, int]): The optional maximum number of seconds to wait for the jobs to be finished
                before returning.<br/>
                If not provided and *wait* is True, the function waits indefinitely.
             **properties (dict[str, any]): A key worded variable length list of user additional arguments
                that will be stored within the `Submission^`. It can be accessed via `Submission.properties^`.

        Returns:
            The created `Submission^` containing the information about the submission.
        """
        submission = _SubmissionManagerFactory._build_manager()._create(
            submittable.id,  # type: ignore
            submittable._ID_PREFIX,  # type: ignore
            getattr(submittable, "config_id", None),
            **properties,
        )
        jobs: List[Job] = []
        tasks = submittable._get_sorted_tasks()
        with cls.lock:
            cls.__logger.debug(f"Acquiring lock to submit {submission.entity_id}.")
            for ts in tasks:
                jobs.extend(
                    cls._lock_dn_output_and_create_job(
                        task,
                        submission.id,
                        submission.entity_id,
                        callbacks=itertools.chain([cls._update_submission_status], callbacks or []),
                        force=force,  # type: ignore
                    )
                    for task in ts
                )
            submission.jobs = jobs  # type: ignore
            cls._orchestrate_job_to_run_or_block(jobs)
        if Config.job_config.is_development:
            cls._check_and_execute_jobs_if_development_mode()
        elif wait:
            cls._wait_until_job_finished(jobs, timeout)
        return submission

    @classmethod
    def submit_task(
        cls,
        task: Task,
        callbacks: Optional[Iterable[Callable]] = None,
        force: bool = False,
        wait: bool = False,
        timeout: Optional[Union[float, int]] = None,
        **properties,
    ) -> Submission:
        """Submit the given `Task^` for an execution.

        Parameters:
             task (Task^): The task to submit for execution.
             callbacks: The optional list of functions that should be executed on job status change.
             force (bool): Enforce execution of the task even if its output data nodes are cached.
             wait (bool): Wait for the orchestrated job created from the task submission to be finished
                in asynchronous mode.
             timeout (Union[float, int]): The optional maximum number of seconds to wait for the job
                to be finished before returning.<br/>
                If not provided and *wait* is True, the function waits indefinitely.
             **properties (dict[str, any]): A key worded variable length list of user additional arguments
                that will be stored within the `Submission^`. It can be accessed via `Submission.properties^`.

        Returns:
            The created `Submission^` containing the information about the submission.
        """
        submission = _SubmissionManagerFactory._build_manager()._create(
            task.id, task._ID_PREFIX, task.config_id, **properties
        )
        submit_id = submission.id
        with cls.lock:
            cls.__logger.debug(f"Acquiring lock to submit task {task.id}.")
            job = cls._lock_dn_output_and_create_job(
                task,
                submit_id,
                submission.entity_id,
                itertools.chain([cls._update_submission_status], callbacks or []),
                force,
            )
            jobs = [job]
            submission.jobs = jobs  # type: ignore
            cls._orchestrate_job_to_run_or_block(jobs)
        if Config.job_config.is_development:
            cls._check_and_execute_jobs_if_development_mode()
        else:
            if wait:
                cls._wait_until_job_finished(job, timeout)
        return submission

    @classmethod
    def _lock_dn_output_and_create_job(
        cls,
        task: Task,
        submit_id: str,
        submit_entity_id: str,
        callbacks: Optional[Iterable[Callable]] = None,
        force: bool = False,
    ) -> Job:
        for dn in task.output.values():
            dn.lock_edit()
        return _JobManagerFactory._build_manager()._create(
            task, itertools.chain([cls._on_status_change], callbacks or []), submit_id, submit_entity_id, force=force
        )

    @classmethod
    def _update_submission_status(cls, job: Job) -> None:
        submission_manager = _SubmissionManagerFactory._build_manager()
        if submission := submission_manager._get(job.submit_id):
            submission_manager._update_submission_status(submission, job)
        else:
            submissions = submission_manager._get_all()
            cls.__logger.error(f"Submission {job.submit_id} not found.")
            msg = "\n--------------------------------------------------------------------------------\n"
            msg += f"Submission {job.submit_id} not found.\n"
            msg += "                              --------------                                    \n"
            msg += "                          Existing submissions                                  \n"
            for s in submissions:
                msg += f"{s.id}\n"
            msg += "--------------------------------------------------------------------------------\n"

            cls.__logger.error(f"Job {job.id} status: {job.status}")

    @classmethod
    def _orchestrate_job_to_run_or_block(cls, jobs: List[Job]) -> None:
        blocked_jobs = []
        pending_jobs = []

        for job in jobs:
            if cls._is_blocked(job):
                job.blocked()
                blocked_jobs.append(job)
            else:
                job.pending()
                pending_jobs.append(job)

        cls.blocked_jobs.extend(blocked_jobs)
        for job in pending_jobs:
            cls.jobs_to_run.put(job)

    @classmethod
    def _wait_until_job_finished(cls, jobs: Union[List[Job], Job], timeout: Optional[Union[float, int]] = None) -> None:
        #  Note: this method should be prefixed by two underscores, but it has only one, so it can be mocked in tests.
        def __check_if_timeout(st, to):
            if to is None:
                return True
            return (datetime.now() - st).seconds < to

        start = datetime.now()
        jobs = list(jobs) if isinstance(jobs, Iterable) else [jobs]
        index = 0
        while __check_if_timeout(start, timeout) and index < len(jobs):
            try:
                if jobs[index]._is_finished():
                    index += 1
                else:
                    sleep(0.5)  # Limit CPU usage
            except Exception:
                pass

    @classmethod
    def _is_blocked(cls, obj: Union[Task, Job]) -> bool:
        """Returns True if the execution of the `Job^` or the `Task^` is blocked by the execution of another `Job^`.

        Parameters:
             obj (Union[Task^, Job^]): The job or task entity to run.

        Returns:
             True if one of its input data nodes is blocked.
        """
        input_data_nodes = obj.task.input.values() if isinstance(obj, Job) else obj.input.values()
        data_manager = _DataManagerFactory._build_manager()
        return any(not data_manager._get(dn.id).is_ready_for_reading for dn in input_data_nodes)

    @staticmethod
    def _unlock_edit_on_jobs_outputs(jobs: Union[Job, List[Job], Set[Job]]) -> None:
        jobs = [jobs] if isinstance(jobs, Job) else jobs
        for job in jobs:
            job._unlock_edit_on_outputs()

    @classmethod
    def _on_status_change(cls, job: Job) -> None:
        if job.is_completed() or job.is_skipped():
            cls.__logger.debug(f"{job.id} has been completed or skipped. Unblocking jobs.")
            cls.__unblock_jobs()
        elif job.is_failed():
            cls._fail_subsequent_jobs(job)

    @classmethod
    def __unblock_jobs(cls) -> None:
        with cls.lock:
            cls.__logger.debug("Acquiring lock to unblock jobs.")
            for job in cls.blocked_jobs:
                if not cls._is_blocked(job):
                    cls.__logger.debug(f"Unblocking job: {job.id}.")
                    job.pending()
                    cls.__logger.debug(f"Removing job {job.id} from the blocked_job list.")
                    cls.__remove_blocked_job(job)
                    cls.__logger.debug(f"Adding job {job.id} to the list of jobs to run.")
                    cls.jobs_to_run.put(job)

    @classmethod
    def __remove_blocked_job(cls, job: Job) -> None:
        try:  # In case the job has been removed from the list of blocked_jobs.
            cls.blocked_jobs.remove(job)
        except Exception:
            cls.__logger.warning(f"{job.id} is not in the blocked list anymore.")

    @classmethod
    def cancel_job(cls, job: Job) -> None:
        if job.is_canceled():
            cls.__logger.info(f"{job.id} has already been canceled.")
        elif job.is_abandoned():
            cls.__logger.info(f"{job.id} has already been abandoned and cannot be canceled.")
        elif job.is_failed():
            cls.__logger.info(f"{job.id} has already failed and cannot be canceled.")
        else:
            with cls.lock:
                cls.__logger.debug(f"Acquiring lock to cancel job {job.id}.")
                to_cancel_or_abandon_jobs = {job}
                to_cancel_or_abandon_jobs.update(cls.__find_subsequent_jobs(job.submit_id, set(job.task.output.keys())))
                cls.__remove_blocked_jobs(to_cancel_or_abandon_jobs)
                cls.__remove_jobs_to_run(to_cancel_or_abandon_jobs)
                cls._cancel_jobs(job.id, to_cancel_or_abandon_jobs)
                cls._unlock_edit_on_jobs_outputs(to_cancel_or_abandon_jobs)

    @classmethod
    def __find_subsequent_jobs(cls, submit_id, output_dn_config_ids: Set) -> Set[Job]:
        next_output_dn_config_ids = set()
        subsequent_jobs = set()
        for job in cls.blocked_jobs:
            job_input_dn_config_ids = job.task.input.keys()
            if job.submit_id == submit_id and len(output_dn_config_ids.intersection(job_input_dn_config_ids)) > 0:
                next_output_dn_config_ids.update(job.task.output.keys())
                subsequent_jobs.update([job])
        if next_output_dn_config_ids:
            subsequent_jobs.update(
                cls.__find_subsequent_jobs(submit_id, output_dn_config_ids=next_output_dn_config_ids)
            )
        return subsequent_jobs

    @classmethod
    def __remove_blocked_jobs(cls, jobs: Set[Job]) -> None:
        for job in jobs:
            cls.__remove_blocked_job(job)

    @classmethod
    def __remove_jobs_to_run(cls, jobs: Set[Job]) -> None:
        new_jobs_to_run: Queue = Queue()
        while not cls.jobs_to_run.empty():
            current_job = cls.jobs_to_run.get()
            if current_job not in jobs:
                new_jobs_to_run.put(current_job)
        cls.jobs_to_run = new_jobs_to_run

    @classmethod
    def _fail_subsequent_jobs(cls, failed_job: Job) -> None:
        with cls.lock:
            cls.__logger.debug("Acquiring lock to fail subsequent jobs.")
            to_fail_or_abandon_jobs = set()
            to_fail_or_abandon_jobs.update(
                cls.__find_subsequent_jobs(failed_job.submit_id, set(failed_job.task.output.keys()))
            )
            for job in to_fail_or_abandon_jobs:
                job.abandoned()
            to_fail_or_abandon_jobs.update([failed_job])
            cls.__remove_blocked_jobs(to_fail_or_abandon_jobs)
            cls.__remove_jobs_to_run(to_fail_or_abandon_jobs)
            cls._unlock_edit_on_jobs_outputs(to_fail_or_abandon_jobs)

    @classmethod
    def _cancel_jobs(cls, job_id_to_cancel: JobId, jobs: Set[Job]) -> None:
        for job in jobs:
            if job.is_running():
                cls.__logger.info(f"{job.id} is running and cannot be canceled.")
            elif job.is_completed():
                cls.__logger.info(f"{job.id} has already been completed and cannot be canceled.")
            elif job.is_skipped():
                cls.__logger.info(f"{job.id} has already been skipped and cannot be canceled.")
            elif job_id_to_cancel == job.id:
                job.canceled()
            else:
                job.abandoned()

    @staticmethod
    def _check_and_execute_jobs_if_development_mode() -> None:
        from ._orchestrator_factory import _OrchestratorFactory

        if dispatcher := _OrchestratorFactory._dispatcher:
            dispatcher._execute_jobs_synchronously()
